import logging

OFF = 60
CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

logger = logging.getLogger('tos')


def set_logger(file_path='TosClient.log', name="tos", level=INFO, format_string=None, log_handler=None):
    global logger
    if not format_string:
        format_string = "%(asctime)s %(name)s [%(levelname)s] %(thread)d : %(message)s"
    logger = logging.getLogger(name)
    logger.setLevel(level)
    fh = log_handler
    if fh is None:
        fh = logging.FileHandler(file_path)
    fh.setLevel(level)
    formatter = logging.Formatter(format_string)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


def get_logger():
    return logger



import datetime
import errno
import functools
import logging
import math
import os.path
import re
import socket
import sys
import threading
import time
import urllib.parse
from hashlib import sha256
from io import StringIO
from urllib.parse import unquote_to_bytes, quote

import crcmod as crcmod
import pytz
import six
from pytz import unicode
from requests.structures import CaseInsensitiveDict
from urllib3.util.connection import allowed_gai_family

from .consts import (DEFAULT_MIMETYPE, GMT_DATE_FORMAT,
                     MAX_PART_NUMBER, MAX_PART_SIZE, MIN_PART_SIZE, CHUNK_SIZE,
                     CLIENT_ENCRYPTION_ALGORITHM, SERVER_ENCRYPTION_ALGORITHM, LAST_MODIFY_TIME_DATE_FORMAT,
                     EMPTY_SHA256_HASH, PAYLOAD_BUFFER, ECS_DATE_FORMAT, SLEEP_BASE_TIME)
from .enum import DataTransferType, ACLType, StorageClassType, MetadataDirectiveType, AzRedundancyType, PermissionType, \
    GranteeType, CannedType
from .exceptions import TosClientError
from .log import get_logger
from .mine_type import TYPES_MAP
from io import StringIO

REGION_MAP = {
    'cn-beijing': 'tos-cn-beijing.volces.com',
    'cn-guangzhou': 'tos-cn-guangzhou.volces.com',
    'cn-shanghai': 'tos-cn-shanghai.volces.com',
    'ap-southeast-1': 'tos-ap-southeast-1.volces.com',
}

S3_REGION_LIST = ['tos-s3-cn-beijing.volces.com',
                  'tos-s3-cn-guangzhou.volces.com',
                  'tos-s3-cn-shanghai.volces.com',
                  'tos-s3-ap-southeast-1.volces.com']

REGION_CONTROL_MAP = {
        "cn-beijing":     "tos-control-cn-beijing.volces.com",
		"cn-guangzhou":   "tos-control-cn-guangzhou.volces.com",
		"cn-shanghai":    "tos-control-cn-shanghai.volces.com",
		"cn-beijing2":    "tos-control-cn-beijing2.volces.com",
		"ap-southeast-1": "tos-control-ap-southeast-1.volces.com",
}


def get_value(kv, key, handler=lambda x: x):
    if key in kv:
        return handler(kv[key])
    else:
        return None


def get_etag(kv):
    return get_value(kv, "ETag", lambda x: x.strip('"'))


def to_bytes(data):
    if six.PY2:
        if isinstance(data, unicode):
            return data.encode('utf-8')
    if six.PY3:
        if isinstance(data, str):
            return data.encode(encoding='utf-8')
    return data


def to_str(data):
    if six.PY2:
        if isinstance(data, unicode):
            return data.encode('utf-8')
    if six.PY3:
        if isinstance(data, bytes):
            return data.decode('utf-8')
    return data


def parse_modify_time_to_utc_datetime(value):
    return datetime.datetime.strptime(value, LAST_MODIFY_TIME_DATE_FORMAT).replace(tzinfo=pytz.utc)


def parse_gmt_time_to_utc_datetime(value):
    return datetime.datetime.strptime(value, GMT_DATE_FORMAT).replace(tzinfo=pytz.utc)

def parse_iso_time_to_utc_datetime(value):
    return datetime.datetime.strptime(value,ECS_DATE_FORMAT).astimezone(pytz.utc)


def get_content_type(key):
    """根据文件名后缀，获取文件类型"""
    ext = os.path.splitext(key)[1].lower()
    return TYPES_MAP[ext] if ext in TYPES_MAP else DEFAULT_MIMETYPE


def init_path(path: str, key: str):
    try_make_file_dir(path)
    if os.path.isdir(path) and key.endswith('/'):
        return True
    return False


def try_make_file_dir(path: str):
    if os.path.isdir(path) or os.path.isfile(path):
        return
    file_dir, file_name = os.path.split(path)
    if file_dir:
        os.makedirs(file_dir, exist_ok=True)


def init_checkpoint_dir(checkpoint_file: str):
    if os.path.isdir(checkpoint_file):
        return checkpoint_file
    file_dir, file_name = os.path.split(checkpoint_file)
    if file_dir:
        os.makedirs(file_dir, exist_ok=True)
        return file_dir
    raise TosClientError('checkpoint_file is invalid')


def _make_range_string(start, last):
    if start is None and last is None:
        return ''

    if start is not None and last is not None and start > last:
        raise TosClientError('invalid range format')

    return 'bytes=' + _range(start, last)


def _range(start, last):
    def to_str(pos):
        if pos is None:
            return ''
        else:
            return str(pos)

    return to_str(start) + '-' + to_str(last)


def to_unicode(data):
    """把输入转换为unicode，要求输入是unicode或者utf-8编码的bytes。"""
    return to_str(data)


def is_utf8_with_trigger(s: bytes):
    i = 0
    while i < len(s):
        now = s[i]
        first = first_zero(s[i])
        if first == 0:
            if s[i] < 32 or s[i] == 127:
                return False
            i = i + 1
        elif 2 <= first <= 6:
            if i + first <= len(s):
                for j in range(i + 1, i + first):
                    if s[j] & 0xC0 != 0x80:
                        return False
                i += first
            else:
                return False
        else:
            return False
    return True


def first_zero(a):
    mask = 0x80
    i = 0
    while mask & a != 0:
        mask = mask >> 1
        i = i + 1
    return i


def generate_http_proxies(ip: str, port: int, user_name: str = None, password: str = None):
    # 未配置ip地址 返回空{}
    if ip is None or len(ip) == 0:
        return {}

    proxy = ""
    if user_name is None or len(user_name) == 0:
        # 返回 "http": "http://{ip}:port",
        proxy = 'http://{0}:{1}'.format(ip, port)

    else:
        # 返回  "http": "http://{user_name}:{password}@{ip}:{port}/"
        proxy = "http://{0}:{1}@{2}:{3}/".format(user_name, password, ip, port)

    return {'http': proxy}


def _make_copy_source(src_bucket, src_key, src_version_id):
    copy_source = {}
    if src_bucket:
        copy_source['Bucket'] = src_bucket

    if src_key:
        copy_source['Key'] = src_key

    if src_version_id:
        copy_source['VersionId'] = src_version_id

    return copy_source


def _make_upload_part_file_content(file, offset, part_size, size):
    """创建 具备offset 和读取长度 file-object
    """
    if offset < 0 or part_size < -1:
        return None
    if offset > size:
        return None

    file.seek(offset, os.SEEK_SET)

    if part_size == -1 or (part_size + offset) >= size:
        return SizeAdapter(file, size - offset, init_offset=offset, can_reset=True)
    else:
        return SizeAdapter(file, part_size, init_offset=offset, can_reset=True)


class SizeAdapter(object):
    def __init__(self, file_object, size, init_offset=None, can_reset=False):
        self.file_object = file_object
        self.size = size
        self.offset = 0
        self.init_offset = init_offset
        self.can_reset = can_reset
        if init_offset:
            self.file_object.seek(init_offset, os.SEEK_SET)

    def read(self, amt=None):
        if self.offset >= self.size:
            return ''

        if (amt is None or amt < 0) or (amt + self.offset >= self.size):
            data = self.file_object.read(self.size - self.offset)
            self.offset = self.size
            return data

        self.offset += amt
        return self.file_object.read(amt)

    @property
    def len(self):
        return self.size

    def reset(self):
        if self.can_reset:
            self.offset = 0
            if self.init_offset is not None:
                self.file_object.seek(self.init_offset, os.SEEK_SET)


def meta_header_encode(query, doseq=False, safe='', encoding=None, errors=None,
                       quote_via=quote):
    headers = {}
    if hasattr(query, "items"):
        query = query.items()
    else:
        # It's a bother at times that strings and string-like objects are
        # sequences.
        try:
            # non-sequence items should not work with len()
            # non-empty strings will fail this
            if len(query) and not isinstance(query[0], tuple):
                raise TypeError
            # Zero-length sequences of all types will get here and succeed,
            # but that's a minor nit.  Since the original implementation
            # allowed empty dicts that type of behavior probably should be
            # preserved for consistency
        except TypeError:
            ty, va, tb = sys.exc_info()
            raise TypeError("not a valid non-string sequence "
                            "or mapping object").with_traceback(tb)

    if not doseq:
        for k, v in query:
            if isinstance(k, bytes):
                k = quote_via(k, safe)
            else:
                k = quote_via(str(k), safe, encoding, errors)

            if isinstance(v, bytes):
                v = quote_via(v, safe)
            else:
                v = quote_via(str(v), safe, encoding, errors)
            headers[k] = v
    else:
        for k, v in query:
            if isinstance(k, bytes):
                k = quote_via(k, safe)
            else:
                k = quote_via(str(k), safe, encoding, errors)

            if isinstance(v, bytes):
                v = quote_via(v, safe)
                headers[k] = v
            elif isinstance(v, str):
                v = quote_via(v, safe, encoding, errors)
                headers[k] = v
            else:
                try:
                    # Is this a sufficient test for sequence-ness?
                    len(v)
                except TypeError:
                    # not a sequence
                    v = quote_via(str(v), safe, encoding, errors)
                    headers[k] = v
                else:
                    # loop over the sequence
                    for elt in v:
                        if isinstance(elt, bytes):
                            elt = quote_via(elt, safe)
                        else:
                            elt = quote_via(str(elt), safe, encoding, errors)
                        headers[k] = v
    return headers


def meta_header_decode(headers):
    decode_headers = {}
    encoding = 'utf-8'
    errors = 'replace'

    for en in headers:
        k = en
        v = headers[en]
        if '%' in k:
            k = unquote_to_bytes(en).decode(encoding, errors)
        if '%' in v:
            v = unquote_to_bytes(headers[en]).decode(encoding, errors)
        decode_headers[k] = v
    return decode_headers


def convert_meta(meta_list, disable_encoding_meta=None):
    if not meta_list:
        return meta_list
    meta = {}
    for m in meta_list:
        meta[m['Key']] = m['Value']
    if disable_encoding_meta:
        return meta
    return meta_header_decode(meta)


def content_disposition_encode(s):
    if not s:
        return s
    content_dispositions = s.split(';')
    result = []
    for content_disposition in content_dispositions:
        if "=" in content_disposition:
            i = content_disposition.index("=")
            if content_disposition[:i].strip().lower() == "filename":
                value = urllib.parse.quote(content_disposition[i + 1:], safe=" \"\'")
                result.append(content_disposition[:i] + "=" + value)
                continue
        result.append(content_disposition)
    return ";".join(result)


def makedir_p(dirpath):
    try:
        os.makedirs(dirpath)
    except os.error as e:
        if e.errno != errno.EEXIST:
            raise


def get_number(m, n):
    return (m + n - 1) // n


def file_bytes_to_read(fileobj):
    current = fileobj.tell()

    fileobj.seek(0, os.SEEK_END)
    end = fileobj.tell()
    fileobj.seek(current, os.SEEK_SET)

    return end - current


def get_parent_directory_from_File(absolute_path_to_file):
    arr = absolute_path_to_file.split(os.sep)[:-1]
    return os.sep.join(arr)


def _cal_progress_callback(progress_callback, consumed_bytes, total_bytes, rw_once_bytes,
                           type: DataTransferType):
    if progress_callback:
        progress_callback(consumed_bytes, total_bytes, rw_once_bytes, type)


def _cal_rate_limiter_callback(rate_limiter, want):
    if rate_limiter:
        while True:
            result = rate_limiter.acquire(want)
            if result.ok:
                return
            time.sleep(result.time_to_wait)


def _cal_crc_callback(crc_callback, content, discard=0):
    if crc_callback:
        crc_callback(content[discard:])


def BaseCallback(object):
    def success():
        pass

    def fail():
        pass

    def abort():
        pass


class _ReaderAdapter(object):
    """
    通过adapter模式，实现进度条监控、客户端限速、上传crc计算
    """

    def __init__(self, data, progress_callback=None, size=None, crc_callback=None,
                 limiter_callback=None, download_operator=False, can_reset=False,
                 init_offset=None):
        self.data = to_bytes(data)
        self.progress_callback = progress_callback
        self.size = size
        self.offset = 0
        self.download_operator = download_operator

        self.crc_callback = crc_callback
        self.limiter_callback = limiter_callback
        if hasattr(data, 'can_reset'):
            self.can_reset = data.can_reset
        else:
            self.can_reset = can_reset
        self.init_offset = init_offset
        if self.init_offset is not None:
            self.data.seek(init_offset, os.SEEK_SET)

    @property
    def len(self):
        return self.size

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        content = self.read(CHUNK_SIZE)
        if content:
            return content
        else:
            raise StopIteration

    def read(self, amt=None):
        if self.offset >= self.size:
            _cal_progress_callback(self.get_progress(), self.size, self.size, 0,
                                   DataTransferType.Data_Transfer_Succeed)
            return to_bytes('')
        if self.offset == 0:
            _cal_progress_callback(self.progress_callback, min(self.offset, self.size), self.size, 0,
                                   DataTransferType.Data_Transfer_Started)
        if amt is None or amt < 0:
            bytes_to_read = self.size - self.offset
        else:
            bytes_to_read = min(amt, self.size - self.offset)

        if isinstance(self.data, bytes):
            content = self.data[self.offset:self.offset + bytes_to_read]
        else:
            content = to_bytes(self.data.read(bytes_to_read))

        self.offset += bytes_to_read

        _cal_progress_callback(self.progress_callback, min(self.offset, self.size), self.size, bytes_to_read,
                               DataTransferType.Data_Transfer_RW)

        _cal_crc_callback(self.crc_callback, content)

        _cal_rate_limiter_callback(self.limiter_callback, bytes_to_read)
        return content

    @property
    def crc(self):
        if self.crc_callback:
            return self.crc_callback.crc
        else:
            return 0

    def reset(self):
        if self.can_reset and self.size != -1:
            self.offset = 0
            if self.crc_callback:
                self.crc_callback = self.crc_callback.reset()
            if self.init_offset is not None:
                self.data.seek(self.init_offset, os.SEEK_SET)
            if isinstance(self.data, _ReaderAdapter) or isinstance(self.data, SizeAdapter):
                self.data.reset()

    def get_progress(self):
        if self.progress_callback:
            return self.progress_callback
        elif isinstance(self.data, _ReaderAdapter):
            return self.data.get_progress()
        return None


# 只对上传时会将网络流包或具备__iter__装成 _IterableAdapter 适配器
# 不支持超时重试、进度条功能
class _IterableAdapter(object):
    def __init__(self, data, progress_callback=None, crc_callback=None, limiter_callback=None, download_operator=False,
                 can_reset=False):
        self.iter = iter(data)
        self.progress_callback = progress_callback
        self.crc_callback = crc_callback
        self.limiter_callback = limiter_callback
        self.download_operator = download_operator
        self.can_reset = False
        self.offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if self.offset == 0:
            _cal_progress_callback(self.progress_callback, self.offset, -1, 0,
                                   DataTransferType.Data_Transfer_Started)
        try:
            content = next(self.iter)
        except Exception as e:
            _cal_progress_callback(self.progress_callback, self.offset, self.offset, 0,
                                   DataTransferType.Data_Transfer_Succeed)
            raise e
        size = len(content)
        _cal_progress_callback(self.progress_callback, self.offset, -1, size,
                               DataTransferType.Data_Transfer_RW)

        _cal_crc_callback(self.crc_callback, content)

        _cal_rate_limiter_callback(self.limiter_callback, size)
        self.offset = self.offset + size
        return content

    def get_progress(self):
        if self.progress_callback:
            return self.progress_callback
        elif isinstance(self.iter, _ReaderAdapter):
            return self.iter.get_progress()
        return None

    @property
    def crc(self):
        if self.crc_callback:
            return self.crc_callback.crc
        else:
            return 0


from .thread_ctx import produce_body


def patch_content(content):
    if isinstance(content, _ReaderAdapter) or (hasattr(content, 'read') and callable(content.read)):
        _read = content.read

        def read(amt=None):
            data = _read(amt)
            produce_body(len(data))
            return data

        content.read = read
    elif isinstance(content, _IterableAdapter):
        _next = content.next

        def next():
            data = _next()
            produce_body(len(data))
            return data

        content.next = next


def init_content(data, can_reset=None, init_offset=None):
    """ 此方法用于将上传数据统一包装为 _ReaderAdapter 或 _IterableAdapter，方便后续统一方式处理超时重试

    @param data: 待包装数据
    @param can_reset: 是否可reset
    @param init_offset: 初始数据偏移
    @return:
    """

    # 说明为 put_object_from_file等方式，以及对流进行包装，调用put_object方式时，无需包装
    if isinstance(data, _ReaderAdapter) or isinstance(data, _IterableAdapter):
        return data

    # 明确可以reset
    if can_reset is True:
        return add_Background_func(data, can_reset=True, init_offset=init_offset)

    # 具备seek、tell方法, 初步判定可以 reset
    if hasattr(data, 'seek') and hasattr(data, 'tell') and init_offset is not None:
        return add_Background_func(data, can_reset=True, init_offset=init_offset)

    # 为SizeAdapter及使用限制结束位的文件上传方式，可reset
    if isinstance(data, SizeAdapter):
        return add_Background_func(data, can_reset=True)

    # 及为非IO对象，(str bytes)等初步判定可reset
    if not (hasattr(data, 'seek') and hasattr(data, 'tell')):
        return add_Background_func(data, can_reset=True)

    if isinstance(data, StringIO):
        data = data.read()
        return add_Background_func(data, can_reset=True)

    # 兜底不可reset
    return add_Background_func(data, can_reset=False)


def add_Background_func(data, can_reset=False, init_offset=None, size=None):
    """
    此方法根据数据的data类型，将date包装为 _IterableAdapter 或 _ReaderAdapter
    1. 尝试将数据转化为bytes
    2. 尝试获取数据的size(len、seek、tell)
        1. 网络流具备seek、tell但未实现因此返回为None
    3. size 为空 但具备 __iter__ 直接封装为_IterableAdapter、通过http chuck方式发送
    4. size 不为空，直接封装为 _ReaderAdapter
    """

    data = to_bytes(data)
    if size is None:
        size = _get_size(data)
    if size is None and hasattr(data, '__iter__'):
        return _IterableAdapter(data, can_reset=False)
    elif size is not None:
        return _ReaderAdapter(data, size=size, can_reset=can_reset, init_offset=init_offset)
    raise TosClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))


def add_progress_listener_func(data, progress_callback, download_operator=False, can_reset=False,
                               init_offset=None, size=None, is_response=False):
    """
    向data添加进度条监控功能，通过adapter模式实现
    """
    data = to_bytes(data)
    if size is None:
        size = _get_size(data)
    if size is None and hasattr(data, '__iter__'):
        return _IterableAdapter(data, progress_callback=progress_callback, can_reset=False)
    elif size is not None or is_response:
        return _ReaderAdapter(data=data, progress_callback=progress_callback, size=size,
                              download_operator=download_operator, can_reset=can_reset, init_offset=init_offset)
    raise TosClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))


def add_rate_limiter_func(data, rate_limiter, size=None, can_reset=False, init_offset=None, is_response=False):
    """
    返回一个适配器，从而在读取、上传 'data'时，能够通过令牌桶算法进行限速
    """
    data = to_bytes(data)
    if size is None:
        size = _get_size(data)
    if size is None and hasattr(data, '__iter__'):
        return _IterableAdapter(data, limiter_callback=rate_limiter, can_reset=False)
    elif size is not None or is_response:
        return _ReaderAdapter(data=data, limiter_callback=rate_limiter, size=size, can_reset=can_reset,
                              init_offset=init_offset)
    raise TosClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))


def add_crc_func(data, init_crc=0, discard=0, size=None, can_reset=False, is_response=False):
    """
    向data中添加crc计算功能，实现上传和下载对象后得到本地对象crc计算值
    """
    data = to_bytes(data)
    if size is None:
        size = _get_size(data)
    if size is None and hasattr(data, '__iter__'):
        return _IterableAdapter(data, crc_callback=Crc64(init_crc), can_reset=False)
    elif size is not None or is_response:
        return _ReaderAdapter(data, size=size, crc_callback=Crc64(init_crc), can_reset=can_reset)
    raise TosClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))


def _get_size(data):
    if isinstance(data, StringIO):
        return len(to_bytes(data.getvalue()[data.tell():]))
    if hasattr(data, '__len__') or hasattr(data, 'len') or (hasattr(data, 'seek') and hasattr(data, 'tell')) or hasattr(
            data, 'read'):
        if hasattr(data, '__len__'):
            return len(data)

        if hasattr(data, 'len'):
            return data.len

        # 具备 seek 和 tell 方法有可能为文件，也可能为网络流， 因此需调用 file_byte_to_read 方法尝试读取长度
        # 抛出异常说明为网络流
        try:
            if hasattr(data, 'seek') and hasattr(data, 'tell'):
                return file_bytes_to_read(data)
        except:
            return None
    return None


class TokenBucketResult(object):
    def __init__(self, ok: bool, time_to_wait: int):
        self.ok = ok
        self.time_to_wait = time_to_wait


class RateLimiter(object):
    """
    令牌桶算法实现
    """

    # rate是令牌发放速度，capacity是桶的⼤⼩
    def __init__(self, rate, capacity):
        self._rate = rate
        self._capacity = capacity
        self._current_amount = 0
        self._last_consume_time = int(time.time())
        self._lock = threading.Lock()

    # want是发送数据需要的令牌数
    def acquire(self, want):
        with self._lock:
            increment = (int(time.time()) - self._last_consume_time) * self._rate  # 计算从上次发送到这次发送，新发放的令牌数量
            self._current_amount = min(
                increment + self._current_amount, self._capacity)  # 令牌数量不能超过桶的容量
            if want > self._current_amount:  # 如果没有⾜够的令牌，则不能发送数据
                time_to_wait = (want - self._current_amount) / self._rate
                return TokenBucketResult(False, time_to_wait)
            self._last_consume_time = int(time.time())
            self._current_amount -= want
            return TokenBucketResult(True, 0)

    def reset(self):
        r = RateLimiter(self._rate, self._capacity)
        return r


class Crc64(object):
    _POLY = 0x142F0E1EBA9EA3693
    _XOROUT = 0XFFFFFFFFFFFFFFFF

    def __init__(self, init_crc=0):
        init_crc = int(init_crc)
        self.init_crc = init_crc
        self.crc64 = crcmod.Crc(self._POLY, initCrc=init_crc, rev=True, xorOut=self._XOROUT)

        self.crc64_combineFun = mkCombineFun(self._POLY, initCrc=init_crc, rev=True, xorOut=self._XOROUT)

    def __call__(self, data):
        self.update(data)

    def update(self, data):
        self.crc64.update(data)

    def combine(self, crc1, crc2, len2):
        crc2 = int(crc2)
        return self.crc64_combineFun(crc1, crc2, len2)

    @property
    def crc(self):
        return self.crc64.crcValue

    def reset(self):
        c = Crc64(init_crc=self.init_crc)
        return c


def check_crc(operation, client_crc, tos_crc, request_id):
    if tos_crc is not None:
        tos_crc = int(tos_crc)
    if client_crc != tos_crc:
        raise TosClientError(
            "Check CRC failed: req_id: {0}, operation: {1}, CRC checksum of client: {2} is mismatch "
            "with tos: {3}".format(request_id, operation, client_crc, tos_crc))


def rename_file(src, dst):
    os.rename(src, dst)


def cal_crc_from_download_parts(parts, init_crc=0):
    client_crc = 0
    crc_obj = Crc64(init_crc)
    for part in parts:
        if not part.part_crc or not part.size:
            return None
        client_crc = crc_obj.combine(client_crc, part.part_crc, part.size)
    return client_crc


def cal_crc_from_upload_parts(parts, init_crc=0):
    client_crc = 0
    crc_obj = Crc64(init_crc)
    for part in parts:
        client_crc = crc_obj.combine(client_crc, part.hash_crc64_ecma, part.part_size)
    return client_crc


def copy_and_verify_length(src, dst, expected_len,
                           chunk_size=64 * 1024,
                           request_id=''):
    num = 0

    while 1:
        buf = src.read(chunk_size)
        if not buf:
            break

        num += len(buf)
        dst.write(buf)

    if num != expected_len:
        raise TosClientError("Some error from read source, request_id:{0}".format(request_id))


is_py3 = (sys.version_info[0] == 3)
if is_py3:
    xrange = range
    long = int
    sys.maxint = sys.maxsize


def mkCombineFun(poly, initCrc=~long(0), rev=True, xorOut=0):
    (sizeBits, initCrc, xorOut) = _verifyParams(poly, initCrc, xorOut)

    mask = (long(1) << sizeBits) - 1
    if rev:
        poly = _bitrev(long(poly) & mask, sizeBits)
    else:
        poly = long(poly) & mask

    if sizeBits == 64:
        fun = _combine64
    else:
        raise NotImplementedError

    def combine_fun(crc1, crc2, len2):
        return fun(poly, initCrc ^ xorOut, rev, xorOut, crc1, crc2, len2)

    return combine_fun


GF2_DIM = 64


def gf2_matrix_square(square, mat):
    for n in xrange(GF2_DIM):
        square[n] = gf2_matrix_times(mat, mat[n])


def gf2_matrix_times(mat, vec):
    summary = 0
    mat_index = 0

    while vec:
        if vec & 1:
            summary ^= mat[mat_index]

        vec >>= 1
        mat_index += 1

    return summary


def _combine64(poly, initCrc, rev, xorOut, crc1, crc2, len2):
    if len2 == 0:
        return crc1

    even = [0] * GF2_DIM
    odd = [0] * GF2_DIM

    crc1 ^= initCrc ^ xorOut

    if (rev):
        # put operator for one zero bit in odd
        odd[0] = poly  # CRC-64 polynomial
        row = 1
        for n in xrange(1, GF2_DIM):
            odd[n] = row
            row <<= 1
    else:
        row = 2
        for n in xrange(0, GF2_DIM - 1):
            odd[n] = row
            row <<= 1
        odd[GF2_DIM - 1] = poly

    gf2_matrix_square(even, odd)

    gf2_matrix_square(odd, even)

    while True:
        gf2_matrix_square(even, odd)
        if len2 & long(1):
            crc1 = gf2_matrix_times(even, crc1)
        len2 >>= 1
        if len2 == 0:
            break

        gf2_matrix_square(odd, even)
        if len2 & long(1):
            crc1 = gf2_matrix_times(odd, crc1)
        len2 >>= 1

        if len2 == 0:
            break

    crc1 ^= crc2

    return crc1


def _verifyPoly(poly):
    msg = 'The degree of the polynomial must be 8, 16, 24, 32 or 64'
    poly = long(poly)  # Use a common representation for all operations
    for n in (8, 16, 24, 32, 64):
        low = long(1) << n
        high = low * 2
        if low <= poly < high:
            return n
    raise ValueError(msg)


def _bitrev(x, n):
    x = long(x)
    y = long(0)
    for i in xrange(n):
        y = (y << 1) | (x & long(1))
        x = x >> 1
    if ((long(1) << n) - 1) <= sys.maxint:
        return int(y)
    return y


def _verifyParams(poly, initCrc, xorOut):
    sizeBits = _verifyPoly(poly)

    mask = (long(1) << sizeBits) - 1

    initCrc = long(initCrc) & mask
    if mask <= sys.maxint:
        initCrc = int(initCrc)

    xorOut = long(xorOut) & mask
    if mask <= sys.maxint:
        xorOut = int(xorOut)

    return (sizeBits, initCrc, xorOut)


def gen_key(host, port):
    return '{}:{}'.format(host, port)


class CacheEntry(object):
    def __init__(self, host, port, ip_list, expire):
        self.host = host
        self.ip_list = ip_list
        self.port = port
        self.expire = expire
        self.immortal = False

    def remove(self, entry):
        if entry in self.ip_list and len(self.ip_list) > 1:
            self.ip_list.remove(entry)

    def copy_ip_list(self):
        return self.ip_list.copy()

    def get_key(self):
        return gen_key(self.host, self.port)


class DnsCacheService(object):
    def __init__(self):
        self.cache = {}
        self.async_started = False
        self.is_shutdown = True
        self.async_lock = threading.Lock()

    def shutdown(self):
        with self.async_lock:
            self.async_started = False
            self.is_shutdown = True

    def async_refresh_cache(self, dns_cache_time, interval=30):
        if self.async_started:
            return False

        with self.async_lock:
            if self.async_started:
                return False

            def _refresh_cache():
                while not self.is_shutdown:
                    try:
                        time.sleep(interval)
                        for k, v in self.cache.copy().items():
                            ip_list = resolve_ip_list(v.host, v.port)
                            if ip_list and len(ip_list) > 0:
                                self.cache[k] = CacheEntry(v.host, v.port, ip_list, time.time() + dns_cache_time)
                            else:
                                v.immortal = True
                    except Exception as ex:
                        get_logger().info('_refresh_cache exception, {}'.format(ex))

            self.async_started = True
            self.is_shutdown = False
            t = threading.Thread(target=_refresh_cache)
            t.setDaemon(True)
            t.start()
            return True

    def get_ip_list(self, host: str, port: int) -> CacheEntry:
        now = int(time.time())
        key = gen_key(host, port)
        if key in self.cache:
            info = self.cache[key]
            if info.immortal or info.expire >= now:
                return info
            self.cache.pop(key)

    def add(self, host, port, ip_list, expire):
        key = gen_key(host, port)
        if key in self.cache:
            return self.cache[key]
        entry = CacheEntry(host, port, ip_list, expire)
        self.cache[key] = entry
        get_logger().info('in-request: add cache address:{}'.format(key))
        return entry


def resolve_ip_list(host, port):
    try:
        family = allowed_gai_family()
        return socket.getaddrinfo(host, port, family, socket.SOCK_STREAM)
    except:
        return None


def check_enum_type(acl=None, storage_class=None, metadata_directive=None, az_redundancy=None,
                    permission=None, grantee=None, canned=None):
    if acl:
        check_acl_type(acl)

    if storage_class:
        check_storage_class_type(storage_class)

    if metadata_directive:
        check_metadata_directive_type(metadata_directive)

    if az_redundancy:
        check_az_redundancy_type(az_redundancy)

    if permission:
        check_permission_type(permission)

    if grantee:
        check_grantee_type(grantee)

    if canned:
        check_canned_type(canned)


def check_acl_type(obj):
    if not isinstance(obj, ACLType):
        raise TosClientError('invalid acl type')


def check_storage_class_type(obj):
    if not isinstance(obj, StorageClassType):
        raise TosClientError('invalid storage class')


def check_metadata_directive_type(obj):
    if not isinstance(obj, MetadataDirectiveType):
        raise TosClientError('invalid metadata directive type')


def check_az_redundancy_type(obj):
    if not isinstance(obj, AzRedundancyType):
        raise TosClientError('invalid az redundancy type')


def check_permission_type(obj):
    if not isinstance(obj, PermissionType):
        raise TosClientError('invalid permission type')


def check_grantee_type(obj):
    if not isinstance(obj, GranteeType):
        raise TosClientError('invalid grantee type')


def check_canned_type(obj):
    if not isinstance(obj, CannedType):
        raise TosClientError('invalid canned type')


def check_part_size(part_size):
    if not (part_size is not None and MIN_PART_SIZE <= part_size <= MAX_PART_SIZE):
        raise TosClientError('invalid part size, the size must be [5242880, 5368709120], size={}'.format(part_size))


def check_part_number(size, part_size):
    number = get_number(size, part_size)
    if number > MAX_PART_NUMBER:
        raise TosClientError('unsupported part number, the maximum is 10000')


def check_client_encryption_algorithm(algorithm):
    if algorithm:
        if algorithm not in CLIENT_ENCRYPTION_ALGORITHM:
            return TosClientError('invalid encryption-decryption algorithm')


def check_server_encryption_algorithm(algorithm):
    if algorithm:
        if algorithm not in SERVER_ENCRYPTION_ALGORITHM:
            return TosClientError('invalid encryption-decryption algorithm')


def is_ip(host):
    p = re.compile('^((25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(25[0-5]|2[0-4]\d|[01]?\d\d?)$')
    if p.match(host):
        return True
    else:
        return False


class LogInfo(object):
    def __init__(self):
        self.start = time.perf_counter()

    def fail(self, func_name, e):
        end = time.perf_counter()
        get_logger().info('after-request: {}  exception: {}, usedTime:{} s'.format(func_name, e, end - self.start))
        raise e

    def success(self, func_name, res):
        end = time.perf_counter()
        get_logger().info(
            'after-request: {} exec httpCode: {}, requestId: {}, usedTime: {} s'.format(func_name, res.status,
                                                                                        res.request_id,
                                                                                        end - self.start))


class MergeProcess(object):
    def __init__(self, process, total_bytes, task_num, consumed_bytes):
        self.process = process
        self.totalBytes = total_bytes
        self.consumed_bytes = consumed_bytes
        if consumed_bytes == 0:
            self.status = DataTransferType.Data_Transfer_Init
        else:
            self.status = DataTransferType.Data_Transfer_RW
        self.taskNum = task_num
        self.count = 0
        self.lock = threading.Lock()

    def __call__(self, consumed_bytes, total_bytes, rw_once_bytes, type: DataTransferType):
        with self.lock:
            if type == DataTransferType.Data_Transfer_Started and self.status == DataTransferType.Data_Transfer_Init:
                self.status = DataTransferType.Data_Transfer_Started
                self.process(0, self.totalBytes, 0, self.status)
            elif type == DataTransferType.Data_Transfer_RW:
                self.status = DataTransferType.Data_Transfer_RW
                self.consumed_bytes += rw_once_bytes
                self.process(self.consumed_bytes, self.totalBytes, rw_once_bytes, DataTransferType.Data_Transfer_RW)

            elif type == DataTransferType.Data_Transfer_Succeed:
                self.count += 1
                if self.count == self.taskNum:
                    self.process(self.consumed_bytes, self.totalBytes, 0, DataTransferType.Data_Transfer_Succeed)


class UploadEventHandler(object):
    def __init__(self, cal_back_fuc, bucket, key, file_path, checkpoint_file, upload_id=None):
        super(UploadEventHandler, self).__init__()
        self.cal_back_fuc = cal_back_fuc
        self.bucket = bucket
        self.key = key
        self.upload_id = upload_id
        self.file_path = file_path
        self.checkpoint_file = checkpoint_file

    def _cal_back_event(self, event_type, e=None, part_info=None):
        self.cal_back_fuc(event_type, e, self.bucket, self.key, self.upload_id, self.file_path,
                          self.checkpoint_file, part_info)

    def __call__(self, *args, **kwargs):
        if self.cal_back_fuc:
            self._cal_back_event(*args, **kwargs)


class DownloadEventHandler(object):
    def __init__(self, cal_back_fuc, bucket, key, version_id, file_path, checkpoint_file):
        super(DownloadEventHandler, self).__init__()
        self.cal_back_fuc = cal_back_fuc
        self.bucket = bucket
        self.key = key
        self.version_id = version_id
        self.file_path = file_path
        self.checkpoint_file = checkpoint_file
        self.temp_file_path = file_path + '.temp'

    def _cal_back_event(self, event_type, e=None, part_info=None):
        self.cal_back_fuc(event_type, e, self.bucket, self.key, self.version_id,
                          self.file_path, self.checkpoint_file, self.temp_file_path, part_info)

    def __call__(self, *args, **kwargs):
        if self.cal_back_fuc:
            self._cal_back_event(*args, **kwargs)


class ResumableCopyObject(object):
    def __init__(self, cal_back_fuc, bucket, key, src_bucket, src_key, src_version_id, checkpoint_file, upload_id=None):
        super(ResumableCopyObject, self).__init__()
        self.cal_back_fuc = cal_back_fuc
        self.bucket = bucket
        self.key = key
        self.upload_id = upload_id
        self.src_bucket = src_bucket
        self.src_key = src_key
        self.src_version_id = src_version_id
        self.checkpoint_file = checkpoint_file

    def _cal_back_event(self, event_type, e=None, part_info=None):
        self.cal_back_fuc(event_type, e, self.bucket, self.key, self.upload_id,
                          self.src_bucket, self.src_key, self.src_version_id, self.checkpoint_file, part_info)

    def __call__(self, *args, **kwargs):
        if self.cal_back_fuc:
            self._cal_back_event(*args, **kwargs)


def _param_to_quoted_query(k, v):
    if v:
        if isinstance(v, list):
            return '&'.join(quote(str(k), '') + '=' + quote(str(i), '') for i in v)
        return quote(str(k), '') + '=' + quote(str(v), '')
    else:
        return quote(k, '/~')


def _if_map(region: str, endpoint: str):
    if region in REGION_MAP and not endpoint:
        return REGION_MAP[region]
    else:
        return endpoint

def _get_control_endpoint(region:str,control_endpoint:str):
    if region in REGION_CONTROL_MAP and not control_endpoint:
        return REGION_CONTROL_MAP[region]
    else:
        return control_endpoint


def _format_endpoint(endpoint):
    if not endpoint.startswith('http://') and not endpoint.startswith('https://'):
        return 'https://' + endpoint
    else:
        return endpoint


def _make_uri(bucket=None, key=None):
    if bucket and not key:
        return '/{0}'.format(bucket)
    if bucket and key:
        return '/{0}/{1}'.format(bucket, key)
    return '/'


def _make_virtual_host_uri(key=None):
    if key:
        return '/{0}'.format(key)
    return '/'


def _get_host(endpoint):
    if endpoint.startswith('http://'):
        return endpoint[7:]
    if endpoint.startswith('https://'):
        return endpoint[8:]
    return endpoint


def _get_scheme(endpoint):
    if endpoint.startswith('http://'):
        return 'http://'
    if endpoint.startswith('https://'):
        return 'https://'
    return 'https://'

def _get_control_host(account_id,endpoint):
    if account_id:
        return account_id+'.'+_get_host(endpoint)
    else:
        return _get_host(endpoint)

def _get_virtual_host(bucket, endpoint):
    if bucket:
        return bucket + '.' + _get_host(endpoint)
    else:
        return _get_host(endpoint)

def _get_vector_host(bucket,account_id,endpoint):
    if bucket and account_id:
        return bucket+'-'+account_id+'.'+_get_host(endpoint)
    else:
        return _get_host(endpoint)


def _cal_content_sha256(data):
    if data and hasattr(data, 'seek'):
        position = data.tell()
        read_chunksize = functools.partial(data.read,
                                           PAYLOAD_BUFFER)
        checksum = sha256()
        for chunk in iter(read_chunksize, b''):
            checksum.update(chunk)
        hex_checksum = checksum.hexdigest()
        data.seek(position)
        return hex_checksum
    elif data:
        return sha256(data).hexdigest()
    else:
        return EMPTY_SHA256_HASH

def _make_control_host_url(host,scheme,account_id=None,key=None):
    url = host
    if account_id and key:
        url = '{0}.{1}/{2}'.format(account_id, host, quote(key, '/~'))
    elif account_id and not key:
        url = '{0}.{1}'.format(account_id, host)
    return _format_endpoint(scheme+url)

def _make_virtual_host_url(host, scheme, bucket=None, key=None,account_id=None):
    url = host
    if bucket and account_id and key:
        url = '{0}-{1}.{2}/{3}'.format(bucket,account_id, host,quote(key, '/~'))
    elif bucket and key:
        url = '{0}.{1}/{2}'.format(bucket, host, quote(key, '/~'))
    elif bucket and account_id and not key:
        url = '{0}-{1}.{2}'.format(bucket,account_id, host)
    elif bucket and not key:
        url = '{0}.{1}'.format(bucket, host)
    elif key:
        url = '{0}/{1}'.format(host, quote(key, '/~'))

    return _format_endpoint(scheme + url)


def is_s3_endpoint(endpoint):
    if endpoint in S3_REGION_LIST:
        return True
    return False


def _build_user_agent(
        base_user_agent,
        user_agent_product_name=None,
        user_agent_soft_name=None,
        user_agent_soft_version=None,
        user_agent_customized_key_values=None,
        undefined="undefined"
):
    """
    构建用户代理(User-Agent)字符串

    参数:
        base_user_agent: 基础User-Agent字符串
        user_agent_product_name: 产品名称
        user_agent_soft_name: 软件名称
        user_agent_soft_version: 软件版本
        user_agent_customized_key_values: 自定义键值对字典
        undefined: 未定义值的替代字符串

    返回:
        构建好的User-Agent字符串
    """
    user_agent = base_user_agent

    # 构建产品/软件信息部分
    if user_agent_product_name is not None or user_agent_soft_name is not None or user_agent_soft_version is not None:
        product_name = user_agent_product_name if user_agent_product_name else undefined
        soft_name = user_agent_soft_name if user_agent_soft_name else undefined
        soft_version = user_agent_soft_version if user_agent_soft_version else undefined

        user_agent += " --{}/{}/{}".format(product_name, soft_name, soft_version)

    # 构建自定义键值对部分
    if user_agent_customized_key_values and isinstance(user_agent_customized_key_values, dict):
        customized_parts = []
        for k, v in user_agent_customized_key_values.items():
            customized_parts.append("{}/{}".format(k, v))

        if customized_parts:
            user_agent += " ({})".format(";".join(customized_parts))

    return user_agent

def _sanitize_dict(d: dict):
    if d:
        for k, v in d.items():
            if isinstance(v, list):
                d[k] = [i if isinstance(i, str) else i.decode() if isinstance(i, bytes) else str(i) for i in v]
            else:
                d[k] = v if isinstance(v, str) else v.decode() if isinstance(v, bytes) else str(v)
    return d

def _to_case_insensitive_dict(headers: dict):
    _sanitize_dict(headers)
    return CaseInsensitiveDict(headers)

def _get_sleep_time(rsp, retry_count):
    sleep_time = SLEEP_BASE_TIME * math.pow(2, retry_count - 1)
    if sleep_time > 60:
        sleep_time = 60
    if rsp and (rsp.status == 429 or rsp.status == 503) and 'retry-after' in rsp.headers:
        try:
            sleep_time = max(int(rsp.headers['retry-after']), sleep_time)
        except Exception as e:
            get_logger().warning('try to parse retry-after from headers error: {}'.format(e))
    return sleep_time

def _validate_account_id(account_id: str):
    """
    Validate that account_id is a non-empty string consisting only of digits.
    Raises TosClientError if validation fails.
    """
    if not account_id:
        raise TosClientError("account_id is required")
    if not isinstance(account_id, str) or not account_id.isdigit():
        raise TosClientError("account_id must be a non-empty numeric string")



# -*- coding: utf-8 -*-
import datetime
import json
import os
import time
import time as tim
import unittest
import string
import random

import crcmod
from pytz import UTC

import tos
from tests.common import TosTestBase, clean_and_delete_bucket
from tos import TosClientV2
from tos.checkpoint import TaskExecutor
from tos.clientv2 import USER_AGENT
from tos.consts import BUCKET_TYPE_HNS, BUCKET_TYPE_FNS
from tos.credential import EnvCredentialsProvider
from tos.enum import ACLType, StorageClassType, RedirectType, StatusType, PermissionType, CannedType, GranteeType, \
    VersioningStatusType, ProtocolType, AzRedundancyType, StorageClassInheritDirectiveType, CertStatus, \
    InventoryFormatType, InventoryFrequencyType, InventoryIncludedObjType
from tos.exceptions import TosClientError, TosServerError
from tos.models2 import CORSRule, Rule, Condition, Redirect, PublicSource, SourceEndpoint, MirrorHeader, \
    BucketLifeCycleRule, BucketLifeCycleExpiration, BucketLifeCycleNoCurrentVersionExpiration, \
    Tag, BucketLifeCycleTransition, \
    BucketLifeCycleNonCurrentVersionTransition, BucketLifeCycleAbortInCompleteMultipartUpload, ReplicationRule, \
    Destination, RedirectAllRequestsTo, IndexDocument, ErrorDocument, RoutingRules, RoutingRule, \
    RoutingRuleCondition, RoutingRuleRedirect, CustomDomainRule, RealTimeLogConfiguration, AccessLogConfiguration, \
    CloudFunctionConfiguration, Filter, FilterKey, FilterRule, RocketMQConfiguration, RocketMQConf, Transform, \
    ReplaceKeyPrefix, FetchHeaderToMetaDataRule, BucketEncryptionRule, ApplyServerSideEncryptionByDefault, \
    BucketLifecycleFilter, NotificationRule, NotificationFilter, NotificationFilterKey, NotificationFilterRule, \
    NotificationDestination, DestinationVeFaaS, DestinationRocketMQ, KV, BucketInventoryConfiguration, \
    InventoryDestination, TOSBucketDestination, InventorySchedule, InventoryFilter, InventoryOptionalFields, \
    AccessControlTranslation, PrivateSource, CommonSourceEndpoint, EndpointCredentialProvider, CredentialProvider

tos.set_logger()


class TestBucket(TosTestBase):

    def test_ua(self):
        assert 'v' in tos.clientv2.USER_AGENT
        client = TosClientV2(self.ak, self.sk, self.endpoint, self.region, enable_crc=True, max_retry_count=2,
                             user_agent_product_name='tos', user_agent_soft_name='crr',
                             user_agent_soft_version='v3.0.0')
        print(client.user_agent)
        print(USER_AGENT + " --tos/crr/v3.0.0")
        assert client.user_agent == USER_AGENT + " --tos/crr/v3.0.0"

        assert 'v' in tos.clientv2.USER_AGENT
        client = TosClientV2(self.ak, self.sk, self.endpoint, self.region, enable_crc=True, max_retry_count=2,
                             user_agent_product_name='tos', user_agent_soft_name='crr',
                             user_agent_soft_version='v3.0.0', user_agent_customized_key_values={'aa':'bb', 'cc':'dd'})
        print(client.user_agent)
        print(USER_AGENT + " --tos/crr/v3.0.0 (aa/bb;cc/dd)")
        assert (client.user_agent == USER_AGENT + " --tos/crr/v3.0.0 (aa/bb;cc/dd)")

    def test_hns_bucket_expires(self):
        bucket_name = self.bucket_name + "hcl"
        self.bucket_delete.append(bucket_name)
        # bucket_name = "sun-eafrofzkzphcl"
        rsp = self.client.create_bucket(bucket_name, bucket_type="hns")
        print(rsp)
        assert rsp.status_code == 200
        rsp = self.client.head_bucket(bucket=bucket_name)
        assert rsp.status_code == 200
        assert rsp.bucket_type == BUCKET_TYPE_HNS

        append_key_0 = "hns/test/0.txt"
        rsp = self.client.append_object(bucket=bucket_name, key=append_key_0, offset=0, content="hello0",
                                        content_length=6,object_expires=3)
        assert rsp.status_code == 200
        offset = rsp.next_append_offset
        assert offset == 6
        head_out = self.client.head_object(bucket=bucket_name,key=append_key_0)
        assert head_out.status_code == 200
        assert head_out.expiration is not None

        append_key = "hns/test/2.txt"
        rsp = self.client.append_object(bucket=bucket_name, key=append_key, offset=0, content_length=0,object_expires=3)
        assert rsp.status_code == 200
        rsp = self.client.head_object(bucket=bucket_name, key=append_key)
        assert rsp.status_code == 200
        assert head_out.expiration is not None
        rsp = self.client.append_object(bucket=bucket_name, key=append_key, offset=0, content="hello1",
                                        content_length=6)
        offset = rsp.next_append_offset
        assert rsp.status_code == 200
        assert offset == 6
        rsp = self.client.head_object(bucket=bucket_name, key=append_key)
        assert rsp.status_code == 200

    def test_hns_bucket(self):
        bucket_name = self.bucket_name + "hcl"
        self.bucket_delete.append(bucket_name)
        # bucket_name = "sun-eafrofzkzphcl"
        rsp = self.client.create_bucket(bucket_name, bucket_type="hns")
        print(rsp)
        assert rsp.status_code == 200
        rsp = self.client.head_bucket(bucket=bucket_name)
        assert rsp.status_code == 200
        assert rsp.bucket_type == BUCKET_TYPE_HNS
        key = "hns/test/1.txt"
        rsp = self.client.put_object(bucket=bucket_name, key=key, content="hello")
        assert rsp.status_code == 200
        # rsp = self.client.head_object(bucket=bucket_name, key=key)

        rsp = self.client.get_file_status(bucket=bucket_name, key=key)
        assert rsp.status_code == 200

        append_key_0 = "hns/test/0.txt"
        rsp = self.client.append_object(bucket=bucket_name, key=append_key_0, offset=0, content="hello0",
                                        content_length=6)
        assert rsp.status_code == 200
        offset = rsp.next_append_offset
        assert offset == 6

        append_key = "hns/test/2.txt"
        try:
            rsp = self.client.append_object(bucket=bucket_name, key=append_key, offset=0, content_length=0, if_match='123')
        except TosServerError as e:
            assert e.status_code == 404
        assert rsp.status_code == 200
        rsp = self.client.append_object(bucket=bucket_name, key=append_key, offset=0, content_length=0)
        assert rsp.status_code == 200
        rsp = self.client.get_object(bucket=bucket_name, key=append_key)
        assert rsp.status_code == 200
        rsp = self.client.append_object(bucket=bucket_name, key=append_key, offset=0, content="hello1",
                                        content_length=6)
        offset = rsp.next_append_offset
        assert rsp.status_code == 200
        assert offset == 6
        rsp = self.client.append_object(bucket=bucket_name, key=append_key, offset=offset, content="hello2")
        offset = rsp.next_append_offset
        assert rsp.status_code == 200
        assert offset == 12
        rsp = self.client.get_object(bucket=bucket_name, key=append_key)
        data = rsp.read().decode('utf-8')
        assert data == 'hello1hello2'

        rsp = self.client.list_buckets(bucket_type=BUCKET_TYPE_HNS)
        assert rsp.status_code == 200
        print(rsp)
        hns_number = len(rsp.buckets)
        for bucket in rsp.buckets:
            assert bucket.bucket_type == BUCKET_TYPE_HNS

        rsp = self.client.list_buckets(bucket_type=BUCKET_TYPE_FNS)
        assert rsp.status_code == 200
        fns_number = len(rsp.buckets)
        for bucket in rsp.buckets:
            assert bucket.bucket_type == BUCKET_TYPE_FNS
        rsp = self.client.list_buckets()
        assert rsp.status_code == 200
        bucket_number = len(rsp.buckets)
        assert bucket_number == hns_number + fns_number

    def test_hns_bucket_is_directory(self):
        # hns 桶
        bucket_name = self.bucket_name + "hns"
        self.bucket_delete.append(bucket_name)
        rsp = self.client.create_bucket(bucket_name, bucket_type="hns")
        assert rsp.status_code == 200
        rsp = self.client.head_bucket(bucket=bucket_name)
        assert rsp.status_code == 200
        assert rsp.bucket_type == BUCKET_TYPE_HNS
        key = "hns/test/1.txt"

        rsp = self.client.put_object(bucket=bucket_name, key=key, content="hello")
        assert rsp.status_code == 200
        rsp = self.client.head_object(bucket=bucket_name, key=key)
        assert rsp.is_directory is False
        rsp = self.client.head_object(bucket=bucket_name, key="hns/")
        assert rsp.is_directory is True
        rsp = self.client.head_object(bucket=bucket_name, key="hns")
        assert rsp.is_directory is True
        rsp = self.client.get_object(bucket=bucket_name, key=key)
        assert rsp.is_directory is False
        rsp = self.client.get_object(bucket=bucket_name, key="hns/")
        assert rsp.is_directory is True
        rsp = self.client.get_object(bucket=bucket_name, key="hns")
        assert rsp.is_directory is True

        rsp = self.client.list_objects_type2(bucket=bucket_name, delimiter='/', prefix='hns/')
        rsp = self.client.list_objects(bucket=bucket_name, delimiter='/', prefix='hns/')

        rsp = self.client.list_buckets()
        # 普通桶
        bucket_name = self.bucket_name + "fns"
        self.bucket_delete.append(bucket_name)
        rsp = self.client.create_bucket(bucket_name)
        assert rsp.status_code == 200
        rsp = self.client.head_bucket(bucket=bucket_name)
        assert rsp.status_code == 200
        assert rsp.bucket_type == BUCKET_TYPE_FNS
        key = "fns/test/1.txt"
        rsp = self.client.put_object(bucket=bucket_name, key=key, content="hello")
        assert rsp.status_code == 200
        rsp = self.client.head_object(bucket=bucket_name, key=key)
        assert rsp.is_directory is False
        with self.assertRaises(TosServerError):
            rsp = self.client.head_object(bucket=bucket_name, key="hns")
        with self.assertRaises(TosServerError):
            rsp = self.client.head_object(bucket=bucket_name, key="hns/")
        rsp = self.client.get_object(bucket=bucket_name, key=key)
        assert rsp.is_directory is False
        with self.assertRaises(TosServerError):
            rsp = self.client.get_object(bucket=bucket_name, key="hns")
        with self.assertRaises(TosServerError):
            rsp = self.client.get_object(bucket=bucket_name, key="hns/")


    def test_get_file_status(self):
        bucket_name = self.bucket_name + "basic"
        self.bucket_delete.append(bucket_name)
        out = self.client.create_bucket(bucket_name)
        assert out.status_code == 200
        key = "hns/test/1.txt"
        content = "hello"
        rsp = self.client.put_object(bucket=bucket_name, key=key, content=content)
        assert rsp.status_code == 200
        rsp = self.client.get_file_status(bucket=bucket_name, key=key)
        assert rsp.status_code == 200
        assert rsp.key == key
        do_crc64 = crcmod.mkCrcFun(0x142F0E1EBA9EA3693, initCrc=0, xorOut=0xffffffffffffffff, rev=True)
        c1 = do_crc64(content.encode())
        assert rsp.crc64 == str(c1)

    def test_bucket(self):
        bucket_name = self.bucket_name + "basic"
        self.bucket_delete.append(bucket_name)
        out = self.client.create_bucket(bucket_name)
        self.assertIsNotNone(out.location)

        self.retry_assert(lambda: self.bucket_name + "basic" in (b.name for b in self.client.list_buckets().buckets))
        head_out = self.client.head_bucket(bucket=bucket_name)
        self.assertIsNotNone(head_out.region)
        self.assertEqual(head_out.storage_class, StorageClassType.Storage_Class_Standard)
        key = 'a.txt'
        self.client.put_object(bucket_name, key=key, content='content')

        with self.assertRaises(TosServerError):
            self.client.delete_bucket(bucket_name)

    def test_bucket_with_storage_class(self):
        bucket_name = self.bucket_name + "storage-class"
        self.client.create_bucket(bucket_name, storage_class=StorageClassType.Storage_Class_Ia)
        self.retry_assert(lambda: bucket_name in (b.name for b in self.client.list_buckets().buckets))

        key = 'a.txt'
        self.client.put_object(bucket_name, key=key, content="content")
        head_bucket_out = self.client.head_bucket(bucket_name)
        self.assertEqual(head_bucket_out.storage_class, StorageClassType.Storage_Class_Ia)
        self.assertIsNotNone(head_bucket_out.region)

        with self.assertRaises(TosServerError):
            self.client.delete_bucket(bucket_name)

        list_objects_out = self.client.list_objects(bucket_name)
        self.assertEqual(1, len(list_objects_out.contents))

        self.assertEqual(list_objects_out.contents[0].storage_class.value, StorageClassType.Storage_Class_Ia.value)

        with self.assertRaises(TosServerError):
            self.client.delete_bucket(bucket_name)

        self.client.delete_object(bucket_name, key)
        self.client.delete_bucket(bucket_name)

        with self.assertRaises(TosServerError):
            self.client.head_bucket(bucket_name + "test")

        bucket_name = self.bucket_name + "multi-az"
        self.client.create_bucket(bucket_name, az_redundancy=AzRedundancyType.Az_Redundancy_Multi_Az)
        self.client.delete_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

    def test_create_bucket_with_illegal_name(self):
        # 测试桶名在3-63个字符内
        with self.assertRaises(TosClientError):
            name = ""
            for i in range(64):
                name = name + '1'
            self.client.create_bucket(name)
        with self.assertRaises(TosClientError):
            self.client.create_bucket("12")

        with self.assertRaises(TosClientError):
            self.client.create_bucket("12A3")
        with self.assertRaises(TosClientError):
            self.client.create_bucket("12_3")

        with self.assertRaises(TosClientError):
            self.client.create_bucket("_123")

        with self.assertRaises(TosClientError):
            self.client.create_bucket("123_")

    def test_list_bucket(self):
        bucket_name = self.bucket_name + "listbucket"
        self.bucket_delete.append(bucket_name)

        self.client.create_bucket(bucket=bucket_name)
        list_out = self.client.list_buckets()
        self.assertTrue(len(list_out.buckets) > 1)
        self.assertTrue(bucket_name in (b.name for b in self.client.list_buckets().buckets))
        self.assertIsNotNone(list_out.owner)
        for bucket in list_out.buckets:
            self.assertIsNotNone(bucket.location)
            self.assertIsNotNone(bucket.name)
            self.assertIsNotNone(bucket.intranet_endpoint)
            self.assertIsNotNone(bucket.extranet_endpoint)
            self.assertIsNotNone(bucket.creation_date)
        self.assertIsNotNone(list_out.owner.id)

    def test_bucket_info(self):
        bucket_name = self.bucket_name + "-bucket-info"
        self.bucket_delete.append(bucket_name)
        with self.assertRaises(TosServerError):
            self.client.head_bucket(bucket_name)

        out = self.client.create_bucket(bucket_name, storage_class=StorageClassType.Storage_Class_Ia)
        self.retry_assert(lambda: bucket_name in (b.name for b in self.client.list_buckets().buckets))

        list_bucket_out = self.client.list_buckets()
        self.assertTrue(len(list_bucket_out.buckets) >= 1)
        self.assertTrue(len(list_bucket_out.owner.id) > 0)
        self.assertTrue(len(list_bucket_out.id2) > 0)
        self.assertTrue(len(list_bucket_out.request_id) > 0)
        self.assertTrue(list_bucket_out.status_code == 200)

        bucket = list_bucket_out.buckets[0]
        self.assertTrue(len(bucket.name) > 0)
        self.assertTrue(len(bucket.location) > 0)
        self.assertTrue(len(bucket.creation_date) > 0)
        self.assertTrue(len(bucket.extranet_endpoint) > 0)
        self.assertTrue(len(bucket.intranet_endpoint) > 0)
        owner = list_bucket_out.owner
        self.assertTrue(len(owner.id) > 0)

    def test_bucket_with_acl(self):
        bucket_name = self.bucket_name + "-acl"
        for acl in ACLType:
            if acl is not ACLType.ACL_Bucket_Owner_Entrusted and acl is not ACLType.ACL_Unknown:
                self.client.create_bucket(bucket_name + acl.value, acl=acl)
                self.client.delete_bucket(bucket_name + acl.value)

    def test_bucket_cors(self):
        bucket_name = self.bucket_name + 'cors'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        cors_rules = []
        cors_rules.append(CORSRule(
            allowed_origins=["example*.com"],
            allowed_methods=['PUT'],
            allowed_headers=['*'],
            expose_headers=['x-tos-test'],
            max_age_seconds=100
        ))
        cors_rules.append(CORSRule(
            allowed_origins=["example*1.com"],
            allowed_methods=['GET'],
            allowed_headers=['*'],
            expose_headers=['*'],
            max_age_seconds=200,
            response_vary=True,
        ))
        out = self.client.put_bucket_cors(bucket=bucket_name, cors_rule=cors_rules)
        self.assertIsNotNone(out.id2)
        self.assertIsNotNone(out.request_id)
        out_get = self.client.get_bucket_cors(bucket=bucket_name)
        self.assertTrue(len(out_get.cors_rules), 2)
        self.assertIsNotNone(out_get.request_id)
        self.assertEqual(out_get.cors_rules[0].max_age_seconds, 100)
        self.assertEqual(out_get.cors_rules[0].expose_headers[0], 'x-tos-test')
        self.assertEqual(out_get.cors_rules[0].allowed_headers[0], '*')
        self.assertEqual(out_get.cors_rules[0].allowed_methods[0], 'PUT')
        self.assertEqual(out_get.cors_rules[0].allowed_origins[0], 'example*.com')

        self.assertEqual(out_get.cors_rules[1].max_age_seconds, 200)
        self.assertEqual(out_get.cors_rules[1].expose_headers[0], '*')
        self.assertEqual(out_get.cors_rules[1].allowed_headers[0], '*')
        self.assertEqual(out_get.cors_rules[1].allowed_methods[0], 'GET')
        self.assertEqual(out_get.cors_rules[1].allowed_origins[0], 'example*1.com')
        self.assertEqual(out_get.cors_rules[1].response_vary, True)

        out_delete = self.client.delete_bucket_cors(bucket=bucket_name)
        self.assertIsNotNone(out_delete.request_id)
        with self.assertRaises(TosServerError):
            self.client.get_bucket_cors(bucket=bucket_name)

    def test_put_bucket_storage_class(self):
        bucket_name = self.bucket_name + 'storage-class'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        put_out = self.client.put_bucket_storage_class(bucket=bucket_name,
                                                       storage_class=StorageClassType.Storage_Class_Ia)
        self.assertIsNotNone(put_out.request_id)
        time.sleep(60)
        out = self.client.head_bucket(bucket=bucket_name)
        self.assertEqual(out.storage_class, StorageClassType.Storage_Class_Ia)

        self.client.put_bucket_storage_class(bucket_name, StorageClassType.Storage_Class_Standard)
        time.sleep(60)
        self.assertEqual(self.client.head_bucket(bucket_name).storage_class, StorageClassType.Storage_Class_Standard)

        # self.client.put_bucket_storage_class(bucket_name, StorageClassType.Storage_Class_Archive_Fr)
        # self.assertEqual(self.client.head_bucket(bucket_name).storage_class, StorageClassType.Storage_Class_Archive_Fr)

    def test_get_location(self):
        bucket_name = self.bucket_name + 'location'
        self.bucket_delete.append(bucket_name)
        self.client.create_bucket(bucket=bucket_name)
        out = self.client.get_bucket_location(bucket=bucket_name)
        self.assertIsNotNone(out.region)
        self.assertIsNotNone(out.extranet_endpoint)
        self.assertIsNotNone(out.intranet_endpoint)

    def test_bucket_private_mirror(self):
        bucket_name = self.bucket_name + 'private-mirror'
        self.client.create_bucket(bucket=bucket_name)
        self.bucket_delete.append(bucket_name)

        rules = []
        rules.append(Rule(
            id='1',
            condition=Condition(http_code=404, allow_host=["example.com"], http_method=["GET", "HEAD"]),
            redirect=Redirect(
                redirect_type=RedirectType.Mirror,
                fetch_source_on_redirect=True,
                private_source=PrivateSource(source_endpoint=CommonSourceEndpoint(
                    primary=[EndpointCredentialProvider(
                        endpoint="https://example.com",
                        bucket_name=bucket_name,
                        credential_provider=CredentialProvider("ServiceRoleBackSourceAccessTOS")
                    )],
                    follower=[EndpointCredentialProvider(
                        endpoint="https://example2.com",
                        bucket_name=bucket_name,
                        credential_provider=CredentialProvider("ServiceRoleBackSourceAccessTOS")
                    )]
                ),),
                pass_query=True,
                follow_redirect=True,
                mirror_header=MirrorHeader(pass_all=True, pass_headers=['aaa', 'bbb'], remove=['xxx', 'xxx'],
                                           set_header=[KV("key1", "value1"), KV("key2", "value2")]),
                transform=Transform(with_key_prefix='prefix', with_key_suffix='suffix',
                                    replace_key_prefix=ReplaceKeyPrefix(key_prefix='prefix1', replace_with='replace')),
                fetch_header_to_meta_data_rules=[FetchHeaderToMetaDataRule(source_header='a', meta_data_suffix='b')],
                fetch_source_on_redirect_with_query=True,
            )
        ))
        put_out = self.client.put_bucket_mirror_back(bucket=bucket_name, rules=rules)
        self.assertIsNotNone(put_out.request_id)
        get_out = self.client.get_bucket_mirror_back(bucket=bucket_name)
        self.assertIsNotNone(get_out.request_id)
        self.assertEqual(len(get_out.rules[0].redirect.private_source.source_endpoint.primary), 1)
        self.assertEqual(get_out.rules[0].redirect.private_source.source_endpoint.primary[0].bucket_name,bucket_name)
        self.assertEqual(get_out.rules[0].redirect.private_source.source_endpoint.primary[0].endpoint, 'https://example.com')
        self.assertEqual(get_out.rules[0].redirect.private_source.source_endpoint.primary[0].credential_provider.role,"ServiceRoleBackSourceAccessTOS")

        self.assertEqual(len(get_out.rules[0].redirect.private_source.source_endpoint.follower), 1)
        self.assertEqual(get_out.rules[0].redirect.private_source.source_endpoint.follower[0].bucket_name, bucket_name)
        self.assertEqual(get_out.rules[0].redirect.private_source.source_endpoint.follower[0].endpoint,
                         'https://example2.com')
        self.assertEqual(get_out.rules[0].redirect.private_source.source_endpoint.follower[0].credential_provider.role,
                         "ServiceRoleBackSourceAccessTOS")


    def test_bucket_mirror(self):
        bucket_name = self.bucket_name + 'mirror'
        self.client.create_bucket(bucket=bucket_name)
        self.bucket_delete.append(bucket_name)
        rules = []
        rules.append(Rule(
            id='1',
            condition=Condition(http_code=404,allow_host=["example.com"],http_method=["GET","HEAD"]),
            redirect=Redirect(
                redirect_type=RedirectType.Mirror,
                fetch_source_on_redirect=True,
                public_source=PublicSource(SourceEndpoint(primary=['http://test.com/obj/tostest/'])),
                pass_query=True,
                follow_redirect=True,
                mirror_header=MirrorHeader(pass_all=True, pass_headers=['aaa', 'bbb'], remove=['xxx', 'xxx'],set_header=[KV("key1", "value1"),KV("key2", "value2")]),
                transform=Transform(with_key_prefix='prefix', with_key_suffix='suffix',
                                    replace_key_prefix=ReplaceKeyPrefix(key_prefix='prefix1', replace_with='replace')),
                fetch_header_to_meta_data_rules=[FetchHeaderToMetaDataRule(source_header='a', meta_data_suffix='b')],
                fetch_source_on_redirect_with_query=True,
            )
        ))
        put_out = self.client.put_bucket_mirror_back(bucket=bucket_name, rules=rules)
        self.assertIsNotNone(put_out.request_id)
        get_out = self.client.get_bucket_mirror_back(bucket=bucket_name)
        self.assertIsNotNone(get_out.request_id)
        self.assertTrue(len(get_out.rules) == 1)
        self.assertEqual(get_out.rules[0].id, '1')
        self.assertEqual(get_out.rules[0].condition.http_code, 404)
        self.assertEqual(get_out.rules[0].redirect.redirect_type, RedirectType.Mirror)
        self.assertEqual(get_out.rules[0].redirect.follow_redirect, True)
        self.assertEqual(get_out.rules[0].redirect.fetch_source_on_redirect, True)
        self.assertEqual(get_out.rules[0].redirect.mirror_header.pass_all, True)
        self.assertEqual(get_out.rules[0].redirect.mirror_header.pass_headers, ['aaa', 'bbb'])
        self.assertEqual(get_out.rules[0].redirect.mirror_header.remove, ['xxx', 'xxx'])
        self.assertEqual(2, len(get_out.rules[0].redirect.mirror_header.set_header))
        self.assertEqual("key1", get_out.rules[0].redirect.mirror_header.set_header[0].key)
        self.assertEqual("value1",get_out.rules[0].redirect.mirror_header.set_header[0].value)
        self.assertEqual("key2", get_out.rules[0].redirect.mirror_header.set_header[1].key)
        self.assertEqual("value2", get_out.rules[0].redirect.mirror_header.set_header[1].value)
        self.assertEqual('GET',get_out.rules[0].condition.http_method[0])
        self.assertEqual('HEAD', get_out.rules[0].condition.http_method[1])
        self.assertEqual(1, len(get_out.rules[0].condition.allow_host))
        self.assertEqual('example.com',get_out.rules[0].condition.allow_host[0])
        self.assertEqual(True, get_out.rules[0].redirect.fetch_source_on_redirect_with_query)
        self.assertEqual(get_out.rules[0].redirect.public_source.source_endpoint.primary,
                         ['http://test.com/obj/tostest/'])
        self.assertEqual(get_out.rules[0].redirect.public_source.fixed_endpoint, None)
        self.assertEqual(get_out.rules[0].redirect.transform.with_key_prefix, 'prefix')
        self.assertEqual(get_out.rules[0].redirect.transform.with_key_suffix, 'suffix')
        self.assertEqual(get_out.rules[0].redirect.transform.replace_key_prefix.key_prefix, 'prefix1')
        self.assertEqual(get_out.rules[0].redirect.transform.replace_key_prefix.replace_with, 'replace')
        self.assertEqual(get_out.rules[0].redirect.fetch_header_to_meta_data_rules[0].source_header, 'a')
        self.assertEqual(get_out.rules[0].redirect.fetch_header_to_meta_data_rules[0].meta_data_suffix, 'b')

        rules = []
        rules.append(Rule(
            id='2',
            condition=Condition(http_code=404),
            redirect=Redirect(
                redirect_type=RedirectType.Async,
                fetch_source_on_redirect=False,
                public_source=PublicSource(SourceEndpoint(primary=['http://test.com/obj/tostest2/'],
                                                          follower=['http://test.com/obj/tostest2/3']),
                                           fixed_endpoint=True),
                pass_query=False,
                follow_redirect=False,
                mirror_header=MirrorHeader(pass_all=False, pass_headers=['aaa2', 'bbb2'], remove=['xxxx', 'xxxx'])
            )
        ))
        put_out = self.client.put_bucket_mirror_back(bucket=bucket_name, rules=rules)
        get_out = self.client.get_bucket_mirror_back(bucket=bucket_name)
        self.assertEqual(get_out.rules[0].id, '2')
        self.assertEqual(get_out.rules[0].condition.http_code, 404)
        self.assertEqual(get_out.rules[0].redirect.redirect_type, RedirectType.Async)
        self.assertEqual(get_out.rules[0].redirect.follow_redirect, None)
        self.assertEqual(get_out.rules[0].redirect.fetch_source_on_redirect, None)
        self.assertEqual(get_out.rules[0].redirect.mirror_header.pass_all, None)
        self.assertEqual(get_out.rules[0].redirect.mirror_header.pass_headers, ['aaa2', 'bbb2'])
        self.assertEqual(get_out.rules[0].redirect.mirror_header.remove, ['xxxx', 'xxxx'])
        self.assertEqual(get_out.rules[0].redirect.public_source.source_endpoint.primary,
                         ['http://test.com/obj/tostest2/'])
        self.assertEqual(get_out.rules[0].redirect.public_source.source_endpoint.follower,
                         ['http://test.com/obj/tostest2/3'])
        self.assertEqual(get_out.rules[0].redirect.public_source.fixed_endpoint, True)

        delete_out = self.client.delete_bucket_mirror_back(bucket=bucket_name)
        self.assertIsNotNone(delete_out.request_id)

        rules = []
        rules.append(Rule(
            id='1',
            condition=Condition(http_code=404, http_method=['GET', 'HEAD']),
            redirect=Redirect(
                redirect_type=RedirectType.Mirror,
                fetch_source_on_redirect=True,
                public_source=PublicSource(SourceEndpoint(primary=['http://test.com/obj/tostest/'])),
                pass_query=True,
                follow_redirect=True,
                mirror_header=MirrorHeader(pass_all=True, pass_headers=['aaa', 'bbb'], remove=['xxx', 'xxx']),
                transform=Transform(with_key_prefix='prefix', with_key_suffix='suffix',
                                    replace_key_prefix=ReplaceKeyPrefix(key_prefix='prefix1', replace_with='replace'))
            )
        ))
        put_out = self.client.put_bucket_mirror_back(bucket=bucket_name, rules=rules)
        self.assertIsNotNone(put_out.request_id)
        get_out = self.client.get_bucket_mirror_back(bucket=bucket_name)
        self.assertIsNotNone(get_out.request_id)
        self.assertTrue(len(get_out.rules) == 1)
        self.assertEqual(get_out.rules[0].id, '1')
        self.assertEqual(get_out.rules[0].condition.http_code, 404)
        self.assertEqual(get_out.rules[0].condition.http_method[0], 'GET')
        self.assertEqual(get_out.rules[0].condition.http_method[1], 'HEAD')
        self.assertEqual(get_out.rules[0].redirect.redirect_type, RedirectType.Mirror)
        self.assertEqual(get_out.rules[0].redirect.follow_redirect, True)
        self.assertEqual(get_out.rules[0].redirect.fetch_source_on_redirect, True)
        self.assertEqual(get_out.rules[0].redirect.mirror_header.pass_all, True)
        self.assertEqual(get_out.rules[0].redirect.mirror_header.pass_headers, ['aaa', 'bbb'])
        self.assertEqual(get_out.rules[0].redirect.mirror_header.remove, ['xxx', 'xxx'])
        self.assertEqual(get_out.rules[0].redirect.public_source.source_endpoint.primary,
                         ['http://test.com/obj/tostest/'])
        self.assertEqual(get_out.rules[0].redirect.public_source.fixed_endpoint, None)
        self.assertEqual(get_out.rules[0].redirect.transform.with_key_prefix, 'prefix')
        self.assertEqual(get_out.rules[0].redirect.transform.with_key_suffix, 'suffix')
        self.assertEqual(get_out.rules[0].redirect.transform.replace_key_prefix.key_prefix, 'prefix1')
        self.assertEqual(get_out.rules[0].redirect.transform.replace_key_prefix.replace_with, 'replace')
        self.client.delete_bucket(bucket=bucket_name)

    def test_bucket_policy(self):
        bucket_name = self.bucket_name + 'policy'
        policy = {
            "Statement": [
                {
                    "Sid": "internal public",
                    "Effect": "Allow",
                    "Action": ["*"],
                    "Principal": "*",
                    "Resource": [
                        "trn:tos:::{}/*".format(bucket_name),
                        "trn:tos:::{}".format(bucket_name),
                    ],
                }
            ]
        }
        policyStr = json.dumps(policy)
        self.client.create_bucket(bucket_name)
        put_out = self.client.put_bucket_policy(bucket_name, policy=policyStr)
        self.assertEqual(put_out.status_code, 204)
        get_out = self.client.get_bucket_policy(bucket_name)
        self.assertEqual(get_out.status_code, 200)
        delete_out = self.client.delete_bucket_policy(bucket_name)
        self.assertEqual(delete_out.status_code, 204)
        self.assertIsNotNone(delete_out.request_id)
        with self.assertRaises(TosServerError):
            self.client.get_bucket_policy(bucket_name)
        self.client.delete_bucket(bucket_name)

    def test_life_cycle(self):
        bucket_name = self.bucket_name + 'lifecycle'
        rules = []
        rules.append(BucketLifeCycleRule(
            id='1',
            prefix='test',
            status=StatusType.Status_Enable,
            # 指定 Bucket的过期属性
            expiration=BucketLifeCycleExpiration(
                date=datetime.datetime(2022, 11, 30),
                # days=70
            ),
            tags=[Tag(key='1', value="2"), Tag('test', 'test')],
        ))
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_bucket_lifecycle(bucket=bucket_name, rules=rules, allow_same_action_overlap=True)
        out = self.client.get_bucket_lifecycle(bucket=bucket_name)
        self.assertEqual(len(out.rules[0].tags), 2)
        self.assertEqual(out.rules[0].tags[0].key, '1')
        self.assertEqual(out.rules[0].tags[0].value, '2')
        self.assertEqual(out.rules[0].tags[1].key, 'test')
        self.assertEqual(out.rules[0].tags[1].value, 'test')
        self.assertEqual(out.allow_same_action_overlap, True)
        self.client.delete_bucket(bucket=bucket_name)

    def test_lifecycle_days(self):
        bucket_name = self.bucket_name + 'lifecycle'
        rules = []
        rules.append(BucketLifeCycleRule(
            id='1',
            prefix='test',
            status=StatusType.Status_Enable,
            # 指定 Bucket的过期属性
            expiration=BucketLifeCycleExpiration(
                days=70
            ),
            no_current_version_expiration=BucketLifeCycleNoCurrentVersionExpiration(
                no_current_days=70
            ),
            abort_in_complete_multipart_upload=BucketLifeCycleAbortInCompleteMultipartUpload(
                days_after_init=10
            ),
            transitions=[BucketLifeCycleTransition(
                days=20,
                storage_class=StorageClassType.Storage_Class_Ia
            )],
            non_current_version_transitions=[BucketLifeCycleNonCurrentVersionTransition(
                storage_class=StorageClassType.Storage_Class_Ia,
                non_current_days=30
            )],
        ))
        rules.append(BucketLifeCycleRule(
            id='2',
            prefix='log',
            status=StatusType.Status_Disable,
            # 指定 Bucket的过期属性
            expiration=BucketLifeCycleExpiration(
                days=60
            ),
            no_current_version_expiration=BucketLifeCycleNoCurrentVersionExpiration(
                no_current_days=60
            ),
            abort_in_complete_multipart_upload=BucketLifeCycleAbortInCompleteMultipartUpload(
                days_after_init=5
            ),
            transitions=[BucketLifeCycleTransition(
                days=10,
                storage_class=StorageClassType.Storage_Class_Ia
            )],
            non_current_version_transitions=[BucketLifeCycleNonCurrentVersionTransition(
                storage_class=StorageClassType.Storage_Class_Ia,
                non_current_days=10
            )],
        ))
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_bucket_lifecycle(bucket=bucket_name, rules=rules)
        out = self.client.get_bucket_lifecycle(bucket=bucket_name)
        self.assertEqual(len(out.rules), 2)
        # 检验 rule1的正确性
        rule1 = out.rules[0]
        self.assertEqual(rule1.id, '1')
        self.assertEqual(rule1.prefix, 'test')
        self.assertEqual(rule1.status, StatusType.Status_Enable)
        self.assertEqual(rule1.expiration.days, 70)
        self.assertEqual(rule1.no_current_version_expiration.no_current_days, 70)
        self.assertEqual(rule1.abort_in_complete_multipart_upload.days_after_init, 10)
        self.assertEqual(rule1.transitions[0].days, 20)
        self.assertEqual(rule1.transitions[0].storage_class, StorageClassType.Storage_Class_Ia)
        self.assertEqual(rule1.non_current_version_transitions[0].non_current_days, 30)
        self.assertEqual(rule1.non_current_version_transitions[0].storage_class, StorageClassType.Storage_Class_Ia)
        self.assertEqual(out.allow_same_action_overlap, None)
        # 校验 rule2的正确性
        rule2 = out.rules[1]
        self.assertEqual(rule2.id, '2')
        self.assertEqual(rule2.prefix, 'log')
        self.assertEqual(rule2.status, StatusType.Status_Disable)
        self.assertEqual(rule2.expiration.days, 60)
        self.assertEqual(rule2.no_current_version_expiration.no_current_days, 60)
        self.assertEqual(rule2.abort_in_complete_multipart_upload.days_after_init, 5)
        self.assertEqual(rule2.transitions[0].days, 10)
        self.assertEqual(rule2.transitions[0].storage_class, StorageClassType.Storage_Class_Ia)
        self.assertEqual(rule2.non_current_version_transitions[0].non_current_days, 10)
        self.assertEqual(rule2.non_current_version_transitions[0].storage_class, StorageClassType.Storage_Class_Ia)
        delete_out = self.client.delete_bucket_lifecycle(bucket_name)
        self.assertIsNotNone(delete_out.request_id)

    def test_lifecycle_date(self):
        bucket_name = self.bucket_name + 'lifecycle'
        rules = []
        rules.append(BucketLifeCycleRule(
            id='1',
            prefix='test',
            status=StatusType.Status_Enable,
            # 指定 Bucket的过期属性
            expiration=BucketLifeCycleExpiration(
                date=datetime.datetime(2022, 11, 30)
            ),
            no_current_version_expiration=BucketLifeCycleNoCurrentVersionExpiration(
                no_current_days=70
            ),
            abort_in_complete_multipart_upload=BucketLifeCycleAbortInCompleteMultipartUpload(
                days_after_init=10
            ),
            transitions=[BucketLifeCycleTransition(
                date=datetime.datetime(2022, 10, 30),
                storage_class=StorageClassType.Storage_Class_Ia
            )],
            non_current_version_transitions=[BucketLifeCycleNonCurrentVersionTransition(
                storage_class=StorageClassType.Storage_Class_Ia,
                non_current_days=30
            )],
        ))
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_bucket_lifecycle(bucket=bucket_name, rules=rules)
        out = self.client.get_bucket_lifecycle(bucket=bucket_name)
        # 检验 rule1的正确性
        rule1 = out.rules[0]
        self.assertEqual(rule1.id, '1')
        self.assertEqual(rule1.prefix, 'test')
        self.assertEqual(rule1.status, StatusType.Status_Enable)
        self.assertEqual(rule1.expiration.date, datetime.datetime(2022, 11, 30, tzinfo=UTC))
        self.assertEqual(rule1.no_current_version_expiration.no_current_days, 70)
        self.assertEqual(rule1.abort_in_complete_multipart_upload.days_after_init, 10)
        self.assertEqual(rule1.transitions[0].date, datetime.datetime(2022, 10, 30, tzinfo=UTC))
        self.assertEqual(rule1.transitions[0].storage_class, StorageClassType.Storage_Class_Ia)
        self.assertEqual(rule1.non_current_version_transitions[0].non_current_days, 30)
        self.assertEqual(rule1.non_current_version_transitions[0].storage_class, StorageClassType.Storage_Class_Ia)

    def test_lifecycle_filter(self):
        bucket_name = self.bucket_name + 'lifecycle'
        rules = []
        rules.append(BucketLifeCycleRule(
            id='1',
            prefix='test',
            status=StatusType.Status_Enable,
            # 指定 Bucket的过期属性
            expiration=BucketLifeCycleExpiration(
                date=datetime.datetime(2022, 9, 30)
            ),
            no_current_version_expiration=BucketLifeCycleNoCurrentVersionExpiration(
                non_current_date=datetime.datetime(2022, 11, 30)
            ),
            non_current_version_transitions=[BucketLifeCycleNonCurrentVersionTransition(
                storage_class=StorageClassType.Storage_Class_Ia,
                non_current_date=datetime.datetime(2022, 10, 30)
            )],
            filter=BucketLifecycleFilter(
                object_size_greater_than=1,
                object_size_less_than=1000,
                greater_than_include_equal=StatusType.Status_Enable,
                less_than_include_equal=StatusType.Status_Disable,
            )
        ))
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_bucket_lifecycle(bucket=bucket_name, rules=rules)
        out = self.client.get_bucket_lifecycle(bucket=bucket_name)
        # 检验 rule1的正确性
        rule1 = out.rules[0]
        self.assertEqual(rule1.id, '1')
        self.assertEqual(rule1.prefix, 'test')
        self.assertEqual(rule1.status, StatusType.Status_Enable)
        self.assertEqual(rule1.expiration.date, datetime.datetime(2022, 9, 30, tzinfo=UTC))
        self.assertEqual(rule1.no_current_version_expiration.non_current_date,
                         datetime.datetime(2022, 11, 30, tzinfo=UTC))
        self.assertEqual(rule1.non_current_version_transitions[0].non_current_date,
                         datetime.datetime(2022, 10, 30, tzinfo=UTC))
        self.assertEqual(rule1.non_current_version_transitions[0].non_current_date,
                         datetime.datetime(2022, 10, 30, tzinfo=UTC))
        self.assertEqual(rule1.non_current_version_transitions[0].storage_class, StorageClassType.Storage_Class_Ia)
        self.assertEqual(rule1.filter.object_size_greater_than, 1)
        self.assertEqual(rule1.filter.object_size_less_than, 1000)
        self.assertEqual(rule1.filter.greater_than_include_equal, StatusType.Status_Enable)
        self.assertEqual(rule1.filter.less_than_include_equal, StatusType.Status_Disable)

    def test_put_bucket_acl(self):
        bucket_name = self.bucket_name + '-acl'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        put_out = self.client.put_bucket_acl(bucket=bucket_name, grant_write='id={}'.format(2))
        self.assertIsNotNone(put_out.request_id)
        get_out = self.client.get_bucket_acl(bucket=bucket_name)
        self.assertIsNotNone(get_out.owner.id)
        self.assertEqual(get_out.grants[0].permission, PermissionType.Permission_Write)
        self.assertEqual(get_out.grants[0].grantee.id, '2')
        self.assertIsNotNone(get_out.grants[0].grantee.type, GranteeType.Grantee_User)

        self.client.put_bucket_acl(bucket_name, acl=ACLType.ACL_Public_Read)
        get_out = self.client.get_bucket_acl(bucket=bucket_name)
        self.assertIsNotNone(get_out.owner.id)
        self.assertEqual(get_out.grants[0].permission, PermissionType.Permission_Read)
        self.assertEqual(get_out.grants[0].grantee.canned, CannedType.Canned_All_Users)

        try:
            self.client.put_bucket_acl(bucket_name+':8080', acl=ACLType.ACL_Public_Read)
        except TosClientError as e:
            self.assertIsNotNone(e)


        self.client.delete_bucket(bucket=bucket_name)

    def test_put_bucket_replication(self):
        bucket_name_src = self.bucket_name + 'replication'
        bucket_name_crr = self.bucket_name + '-crr'
        bucket_name_crr_2 = self.bucket_name + '-crr2'
        bucket_name_crr_3 = self.bucket_name + '-crr3'
        self.client.create_bucket(bucket_name_src)
        self.client2.create_bucket(bucket_name_crr)
        self.client2.create_bucket(bucket_name_crr_2)
        self.client2.create_bucket(bucket_name_crr_3)
        self.bucket_delete.append(bucket_name_src)
        self.bucket_delete.append(bucket_name_crr)
        self.bucket_delete.append(bucket_name_crr_2)
        self.bucket_delete.append(bucket_name_crr_3)
        rules = []
        rules.append(ReplicationRule(id='1',
                                     status=StatusType.Status_Enable,
                                     prefix_set=['prefix1', 'prefix2'],
                                     destination=Destination(bucket=bucket_name_crr, location=self.region2,
                                                             storage_class=StorageClassType.Storage_Class_Ia,
                                                             storage_class_inherit_directive=StorageClassInheritDirectiveType.Storage_Class_ID_Source_Object),
                                     historical_object_replication=StatusType.Status_Enable))

        rules.append(ReplicationRule(id='2',
                                     status=StatusType.Status_Enable,
                                     prefix_set=['prefix3', 'prefix4'],
                                     destination=Destination(bucket=bucket_name_crr_2, location=self.region2,
                                                             storage_class=StorageClassType.Storage_Class_Standard,
                                                             storage_class_inherit_directive=StorageClassInheritDirectiveType.Storage_Class_ID_Source_Object),
                                     historical_object_replication=StatusType.Status_Disable))

        rules.append(ReplicationRule(id='3',
                                     status=StatusType.Status_Enable,
                                     prefix_set=['prefix5'],
                                     destination=Destination(bucket=bucket_name_crr_3, location=self.region2,
                                                             storage_class=StorageClassType.Storage_Class_Ia,
                                                             storage_class_inherit_directive=StorageClassInheritDirectiveType.Storage_Class_ID_Source_Object),
                                     historical_object_replication=StatusType.Status_Disable,tags=[Tag("key1","value1"),Tag("key2","value2")],
                                     access_control_translation=AccessControlTranslation("BucketOwnerEntrusted")))

        put_out = self.client.put_bucket_replication(bucket_name_src, role='ServiceRoleforReplicationAccessTOS',
                                                     rules=rules)
        self.assertIsNotNone(put_out.request_id)
        out = self.client.get_bucket_replication(bucket_name_src, '1')
        self.assertIsNotNone(out.request_id)
        self.assertTrue(len(out.rules) == 1)
        self.assertEqual(out.rules[0].id, '1')
        self.assertEqual(out.rules[0].prefix_set, ['prefix1', 'prefix2'])
        self.assertEqual(out.rules[0].destination.bucket, bucket_name_crr)
        self.assertEqual(out.rules[0].destination.location, self.region2)
        self.assertEqual(out.rules[0].destination.storage_class, StorageClassType.Storage_Class_Ia)
        self.assertEqual(out.rules[0].destination.storage_class_inherit_directive,
                         StorageClassInheritDirectiveType.Storage_Class_ID_Source_Object)
        self.assertEqual(out.rules[0].historical_object_replication, StatusType.Status_Enable)
        self.assertIsNotNone(out.rules[0].progress)
        self.assertEqual(out.rules[0].progress.historical_object, 0.0)
        out = self.client.get_bucket_replication(bucket_name_src, '2')
        self.assertIsNotNone(out.request_id)
        self.assertTrue(len(out.rules) == 1)
        self.assertEqual(out.rules[0].id, '2')
        self.assertEqual(out.rules[0].prefix_set, ['prefix3', 'prefix4'])
        self.assertEqual(out.rules[0].destination.bucket, bucket_name_crr_2)
        self.assertEqual(out.rules[0].destination.location, self.region2)
        self.assertEqual(out.rules[0].destination.storage_class, StorageClassType.Storage_Class_Standard)
        self.assertEqual(out.rules[0].destination.storage_class_inherit_directive,
                         StorageClassInheritDirectiveType.Storage_Class_ID_Source_Object)
        self.assertEqual(out.rules[0].historical_object_replication, StatusType.Status_Disable)
        self.assertIsNotNone(out.rules[0].progress)
        self.assertEqual(out.rules[0].progress.historical_object, 0.0)

        out = self.client.get_bucket_replication(bucket_name_src, '3')
        self.assertEqual(out.rules[0].id, '3')
        self.assertEqual(out.rules[0].prefix_set, ['prefix5'])
        self.assertEqual(out.rules[0].destination.bucket, bucket_name_crr_3)
        self.assertEqual(out.rules[0].destination.location, self.region2)
        self.assertEqual(out.rules[0].destination.storage_class, StorageClassType.Storage_Class_Ia)
        self.assertEqual(out.rules[0].destination.storage_class_inherit_directive,
                         StorageClassInheritDirectiveType.Storage_Class_ID_Source_Object)
        self.assertEqual(out.rules[0].historical_object_replication, StatusType.Status_Disable)
        self.assertIsNotNone(out.rules[0].progress)
        self.assertEqual(out.rules[0].progress.historical_object, 0.0)
        self.assertEqual(out.rules[0].access_control_translation.owner['Owner'], 'BucketOwnerEntrusted')
        self.assertEqual(len(out.rules[0].tags), 2)
        self.assertEqual(out.rules[0].tags[0].key, 'key1')
        self.assertEqual(out.rules[0].tags[1].key, 'key2')

        out = self.client.get_bucket_replication(bucket_name_src)
        self.assertEqual(len(out.rules), 3)

        out = self.client.delete_bucket_replication(bucket_name_src)
        self.assertIsNotNone(out.request_id)

        with self.assertRaises(TosServerError):
            self.client.get_bucket_replication(bucket_name_src)

    def test_bucket_version(self):
        bucket_name = self.bucket_name + 'version'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        out = self.client.put_bucket_versioning(bucket_name, VersioningStatusType.Versioning_Status_Enabled)
        self.assertIsNotNone(out.request_id)
        time.sleep(30)
        out = self.client.get_bucket_version(bucket_name)
        self.assertIsNotNone(out.request_id)
        self.assertEqual(VersioningStatusType.Versioning_Status_Enabled, out.status)

        self.client.put_bucket_versioning(bucket_name, VersioningStatusType.Versioning_Status_Suspended)
        out = self.client.get_bucket_version(bucket_name)
        time.sleep(30)
        self.assertEqual(out.status, VersioningStatusType.Versioning_Status_Suspended)

    def test_bucket_website(self):
        bucket_name = self.bucket_name + 'website'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        rule1 = RoutingRule(condition=RoutingRuleCondition(key_prefix_equals='prefix'),
                            redirect=RoutingRuleRedirect(protocol=ProtocolType.Protocol_Http, host_name='test2.name',
                                                         replace_key_with='replace_key_with',
                                                         http_redirect_code=302))
        rule2 = RoutingRule(condition=RoutingRuleCondition(http_error_code_returned_equals=403),
                            redirect=RoutingRuleRedirect(protocol=ProtocolType.Protocol_Https, host_name='test3.name',
                                                         replace_key_prefix_with='replace_prefix2',
                                                         http_redirect_code=301))
        redirect_all = RedirectAllRequestsTo('test.com', 'http')
        index_document = IndexDocument('index.html', forbidden_sub_dir=True)
        error_document = ErrorDocument('error.html')
        routing_rules = RoutingRules([rule1, rule2])
        out = self.client.put_bucket_website(bucket=bucket_name, redirect_all_requests_to=redirect_all)
        self.assertIsNotNone(out.request_id)
        out = self.client.get_bucket_website(bucket=bucket_name)
        self.assertEqual(out.redirect_all_requests_to.protocol, 'http')
        self.assertEqual(out.redirect_all_requests_to.host_name, 'test.com')

        self.client.put_bucket_website(bucket_name, index_document=index_document, error_document=error_document,
                                       routing_rules=routing_rules)

        out = self.client.get_bucket_website(bucket_name)
        self.assertEqual(out.index_document.suffix, 'index.html')
        self.assertEqual(out.index_document.forbidden_sub_dir, True)

        self.assertEqual(out.error_document.key, 'error.html')
        self.assertEqual(len(out.routing_rules), 2)
        self.assertEqual(out.routing_rules[0].condition.key_prefix_equals, 'prefix')
        self.assertEqual(out.routing_rules[0].condition.http_error_code_returned_equals, None)
        self.assertEqual(out.routing_rules[0].redirect.host_name, 'test2.name')
        self.assertEqual(out.routing_rules[0].redirect.protocol, ProtocolType.Protocol_Http)
        self.assertEqual(out.routing_rules[0].redirect.http_redirect_code, 302)
        self.assertEqual(out.routing_rules[0].redirect.replace_key_with, 'replace_key_with')

        self.assertEqual(out.routing_rules[1].condition.http_error_code_returned_equals, 403)
        self.assertEqual(out.routing_rules[1].condition.key_prefix_equals, None)
        self.assertEqual(out.routing_rules[1].redirect.host_name, 'test3.name')
        self.assertEqual(out.routing_rules[1].redirect.protocol, ProtocolType.Protocol_Https)
        self.assertEqual(out.routing_rules[1].redirect.http_redirect_code, 301)
        self.assertEqual(out.routing_rules[1].redirect.replace_key_prefix_with, 'replace_prefix2')

        delete_out = self.client.delete_bucket_website(bucket_name)
        self.assertIsNotNone(delete_out.request_id)
        with self.assertRaises(TosServerError):
            out = self.client.get_bucket_website(bucket_name)

    # def test_put_bucket_custom_domain(self):
    #     bucket_name = self.bucket_name + '-custom-domain1'
    #     self.client2.create_bucket(bucket_name)
    #     self.bucket_delete.append(bucket_name)
    #     domain = CustomDomainRule(domain='example22.test.com', forbidden=True, forbidden_reason='test')
    #     domain2 = CustomDomainRule(domain='example33.test.com', forbidden=False, forbidden_reason='test2',
    #                                cert_status=CertStatus.Cert_Status_Bound)
    #     self.client2.put_bucket_custom_domain(bucket_name, domain)
    #     self.client2.put_bucket_custom_domain(bucket_name, domain2)
    #
    #     list_out = self.client2.List_bucket_custom_domain(bucket_name)
    #     self.assertEqual(list_out.rules[0].domain, 'example2.test.com')
    #     self.assertEqual(list_out.rules[1].domain, 'example3.test.com')
    #     out = self.client2.delete_bucket_custom_domain(bucket_name, domain='example2.test.com')
    #     self.assertIsNotNone(out.request_id)
    #     list_out = self.client2.List_bucket_custom_domain(bucket_name)
    #     self.assertEqual(len(list_out.rules), 1)
    #     self.client2.delete_bucket_custom_domain(bucket_name, domain='example3.test.com')
    #     with self.assertRaises(TosServerError):
    #         self.client2.List_bucket_custom_domain(bucket_name)

    def test_put_bucket_real_time_log(self):
        bucket_name = self.bucket_name + '-real-time-log'
        self.client2.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        config = RealTimeLogConfiguration(role='TOSLogArchiveTLSRole',
                                          configuration=AccessLogConfiguration(use_service_topic=True))
        out = self.client2.put_bucket_real_time_log(bucket_name, config)
        self.assertIsNotNone(out.request_id)
        get_out = self.client2.get_bucket_real_time_log(bucket_name)
        self.assertIsNotNone(get_out.request_id)
        self.assertEqual(get_out.configuration.role, 'TOSLogArchiveTLSRole')
        self.assertEqual(get_out.configuration.configuration.use_service_topic, True)
        self.assertIsNotNone(get_out.configuration.configuration.use_service_topic)
        self.assertIsNotNone(get_out.configuration.configuration.tls_project_id)
        delete_out = self.client2.delete_bucket_real_time_log(bucket_name)
        self.assertIsNotNone(delete_out.request_id)
        with self.assertRaises(TosServerError):
            self.client2.get_bucket_real_time_log(bucket_name)

    def test_bucket_tagging(self):
        bucket_name = self.bucket_name  # + '-bucket-tagging'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        tag_set = [Tag(
            key='1',
            value='2'
        ), Tag(
            key='3',
            value='4'
        )]
        self.client.put_bucket_tagging(bucket_name, tag_set)
        out = self.client.get_bucket_tagging(bucket_name)
        self.assertIsNotNone(out.request_id)
        self.assertEqual(len(out.tag_set), 2)
        self.assertEqual(out.tag_set[0].key, tag_set[0].key)
        self.assertEqual(out.tag_set[0].value, tag_set[0].value)
        self.assertEqual(out.tag_set[1].key, tag_set[1].key)
        self.assertEqual(out.tag_set[1].value, tag_set[1].value)
        delete_out = self.client.delete_bucket_tagging(bucket_name)
        self.assertIsNotNone(delete_out.request_id)
        with self.assertRaises(TosServerError):
            self.client.get_bucket_tagging(bucket_name)

    def test_bucket_project_name(self):
        bucket_name = self.bucket_name + "project-name"
        self.bucket_delete.append(bucket_name)

        project_name = 'default'
        self.client.create_bucket(bucket=bucket_name, project_name=project_name)
        head_out = self.client.head_bucket(bucket_name)
        self.assertEqual(head_out.project_name, project_name)
        list_out = self.client.list_buckets(project_name=project_name)
        self.assertTrue(len(list_out.buckets) > 1)
        self.assertTrue(bucket_name in (b.name for b in list_out.buckets))
        for bucket in list_out.buckets:
            self.assertEqual(bucket.project_name, project_name)

    def test_bucket_encryption(self):
        bucket_name = self.bucket_name + "-encryption"
        endpoint = "https://{}".format(_get_clean_endpoint(self.endpoint))
        https_client = TosClientV2(self.ak, self.sk, endpoint, self.region, enable_crc=True, max_retry_count=2)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        https_client.put_bucket_encryption(bucket_name, BucketEncryptionRule(
            apply_server_side_encryption_by_default=ApplyServerSideEncryptionByDefault(
                sse_algorithm="kms",
                kms_master_key_id="123"
            )
        ))
        get_out = self.client.get_bucket_encryption(bucket_name)
        self.assertEqual(get_out.rule.apply_server_side_encryption_by_default.sse_algorithm, "kms")
        self.assertEqual(get_out.rule.apply_server_side_encryption_by_default.kms_master_key_id, "123")
        self.client.delete_bucket_encryption(bucket_name)
        with self.assertRaises(TosServerError):
            self.client.get_bucket_encryption(bucket_name)
    # UT所用实例不可用
    # def test_bucket_notification_type2(self):
    #     bucket_name = self.bucket_name + "-notification-type2"
    #     self.client.create_bucket(bucket_name)
    #     self.bucket_delete.append(bucket_name)
    #
    #     rules = [
    #         NotificationRule(
    #             rule_id="test1",
    #             events=["tos:ObjectCreated:Post", "tos:ObjectCreated:Origin"],
    #             filter=NotificationFilter(
    #                 tos_key=NotificationFilterKey(
    #                     filter_rules=[
    #                         NotificationFilterRule(name="prefix", value="test-")
    #                     ]
    #                 )
    #             ),
    #             destination=NotificationDestination(
    #                 ve_faas=[DestinationVeFaaS(function_id=self.cloud_function)],
    #                 rocket_mq=[
    #                     DestinationRocketMQ(
    #                         role="trn:iam::{}:role/{}".format(self.account_id, self.mq_role_name),
    #                         instance_id=self.mq_instance_id,
    #                         topic="SDK",
    #                         access_key_id=self.mq_access_key_id
    #                     )
    #                 ]
    #             )
    #         )
    #     ]
    #     self.client.put_bucket_notification_type2(bucket_name, rules)
    #     out = self.client.get_bucket_notification_type2(bucket_name)
    #
    #     self.assertTrue(out.version != '')
    #     self.assertEqual(len(out.rules), 1)
    #     self.assertEqual(out.rules[0].rule_id, rules[0].rule_id)
    #     self.assertEqual(out.rules[0].events, rules[0].events)
    #     self.assertEqual(out.rules[0].filter.tos_key.filter_rules[0].name, rules[0].filter.tos_key.filter_rules[0].name)
    #     self.assertEqual(out.rules[0].filter.tos_key.filter_rules[0].value,
    #                      rules[0].filter.tos_key.filter_rules[0].value)
    #     self.assertEqual(out.rules[0].destination.ve_faas[0].function_id, rules[0].destination.ve_faas[0].function_id)
    #     self.assertEqual(out.rules[0].destination.rocket_mq[0].role, rules[0].destination.rocket_mq[0].role)
    #     self.assertEqual(out.rules[0].destination.rocket_mq[0].topic, rules[0].destination.rocket_mq[0].topic)
    #     self.assertEqual(out.rules[0].destination.rocket_mq[0].access_key_id,
    #                      rules[0].destination.rocket_mq[0].access_key_id)
    #     self.assertEqual(out.rules[0].destination.rocket_mq[0].instance_id,
    #                      rules[0].destination.rocket_mq[0].instance_id)

    def test_bucket_inventory(self):
        bucket_name = self.bucket_name + '-inventory'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        inventory_id = "py-sdk-test"
        bucket_inventory_configuration = BucketInventoryConfiguration(
            inventory_id=inventory_id,
            is_enabled=True,
            destination=InventoryDestination(tos_bucket_destination=TOSBucketDestination(
                format=InventoryFormatType.InventoryFormatCsv,
                account_id=self.account_id,
                role="TosArchiveTOSInventory",
                bucket=bucket_name)),
            inventory_filter=InventoryFilter(prefix="prefix1"),
            schedule=InventorySchedule(frequency=InventoryFrequencyType.InventoryFrequencyTypeDaily),
            included_object_versions=InventoryIncludedObjType.InventoryIncludedObjTypeCurrent,

        )
        resp = self.client.put_bucket_inventory(bucket_name,bucket_inventory_configuration)
        self.assertEqual(resp.status_code,200)

        inventory_id2 = "py-sdk-test2"
        bucket_inventory_configuration = BucketInventoryConfiguration(
            inventory_id=inventory_id2,
            is_enabled=False,
            destination=InventoryDestination(tos_bucket_destination=TOSBucketDestination(
                format=InventoryFormatType.InventoryFormatCsv,
                account_id=self.account_id,
                role="TosArchiveTOSInventory",
                bucket=bucket_name)),
            schedule=InventorySchedule(frequency=InventoryFrequencyType.InventoryFrequencyTypeWeekly),
            included_object_versions=InventoryIncludedObjType.InventoryIncludedObjTypeAll,
            inventory_filter=InventoryFilter(prefix="prefix2"),
            optional_fields=InventoryOptionalFields(["Size","CRC64"])
        )
        resp = self.client.put_bucket_inventory(bucket_name, bucket_inventory_configuration)
        self.assertEqual(resp.status_code, 200)


        resp = self.client.get_bucket_inventory(bucket_name,inventory_id=inventory_id)
        self.assertEqual(resp.bucket_inventory_configuration.inventory_id, inventory_id)
        self.assertEqual(resp.bucket_inventory_configuration.is_enabled, True)

        resp = self.client.get_bucket_inventory(bucket_name, inventory_id=inventory_id2)
        self.assertEqual(resp.bucket_inventory_configuration.inventory_id, inventory_id2)
        self.assertEqual(resp.bucket_inventory_configuration.is_enabled, False)
        self.assertEqual(resp.bucket_inventory_configuration.destination.tos_bucket_destination.format, InventoryFormatType.InventoryFormatCsv)
        self.assertEqual(resp.bucket_inventory_configuration.destination.tos_bucket_destination.account_id, self.account_id)
        self.assertEqual(resp.bucket_inventory_configuration.destination.tos_bucket_destination.role, "TosArchiveTOSInventory")
        self.assertEqual(resp.bucket_inventory_configuration.destination.tos_bucket_destination.bucket, bucket_name)
        self.assertEqual(resp.bucket_inventory_configuration.schedule.frequency, InventoryFrequencyType.InventoryFrequencyTypeWeekly)
        self.assertEqual(resp.bucket_inventory_configuration.included_object_versions, InventoryIncludedObjType.InventoryIncludedObjTypeAll)
        self.assertEqual(resp.bucket_inventory_configuration.inventory_filter.prefix, "prefix2")
        self.assertEqual(resp.bucket_inventory_configuration.optional_fields.fields, ["Size","CRC64"])

        resp = self.client.list_bucket_inventory(bucket_name)
        self.assertEqual(len(resp.configurations), 2)
        bucket_inventory_configuration = resp.configurations[1]
        self.assertEqual(bucket_inventory_configuration.inventory_id, inventory_id2)
        self.assertEqual(bucket_inventory_configuration.is_enabled, False)
        self.assertEqual(bucket_inventory_configuration.destination.tos_bucket_destination.format,
                         InventoryFormatType.InventoryFormatCsv)
        self.assertEqual(bucket_inventory_configuration.destination.tos_bucket_destination.account_id,
                         self.account_id)
        self.assertEqual(bucket_inventory_configuration.destination.tos_bucket_destination.role,
                         "TosArchiveTOSInventory")
        self.assertEqual(bucket_inventory_configuration.destination.tos_bucket_destination.bucket, bucket_name)
        self.assertEqual(bucket_inventory_configuration.schedule.frequency,
                         InventoryFrequencyType.InventoryFrequencyTypeWeekly)
        self.assertEqual(bucket_inventory_configuration.included_object_versions,
                         InventoryIncludedObjType.InventoryIncludedObjTypeAll)
        self.assertEqual(bucket_inventory_configuration.inventory_filter.prefix, "prefix2")
        self.assertEqual(bucket_inventory_configuration.optional_fields.fields, ["Size", "CRC64"])

        self.client.delete_bucket_inventory(bucket_name,inventory_id=inventory_id)
        self.client.delete_bucket_inventory(bucket_name, inventory_id=inventory_id2)

        try:
            self.client.get_bucket_inventory(bucket_name, inventory_id=inventory_id)
        except TosServerError as e:
            self.assertEqual(e.status_code,404)

        try:
            self.client.get_bucket_inventory(bucket_name, inventory_id=inventory_id2)
        except TosServerError as e:
            self.assertEqual(e.status_code, 404)

    def test_list_bucket_inventory(self):
        bucket_name = self.bucket_name + '-inventory'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        for i in range(102):
            inventory_id = "py-sdk-test"+str(i)
            bucket_inventory_configuration = BucketInventoryConfiguration(
                inventory_id=inventory_id,
                is_enabled=True,
                destination=InventoryDestination(tos_bucket_destination=TOSBucketDestination(
                    format=InventoryFormatType.InventoryFormatCsv,
                    account_id=self.account_id,
                    role="TosArchiveTOSInventory",
                    bucket=bucket_name)),
                inventory_filter=InventoryFilter(prefix="prefix"+''.join(random.choice(string.ascii_lowercase) for i in range(5))),
                schedule=InventorySchedule(frequency=InventoryFrequencyType.InventoryFrequencyTypeDaily),
                included_object_versions=InventoryIncludedObjType.InventoryIncludedObjTypeCurrent,

            )
            resp = self.client.put_bucket_inventory(bucket_name, bucket_inventory_configuration)
            self.assertEqual(resp.status_code, 200)

        resp = self.client.list_bucket_inventory(bucket_name)
        self.assertEqual(len(resp.configurations), 100)

        resp = self.client.list_bucket_inventory(bucket_name,continuation_token=resp.next_continuation_token)
        self.assertEqual(len(resp.configurations), 1)



    def test_bucket_notification_type2(self):
        bucket_name = self.bucket_name + "-notification-type2"
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        rules = [
            NotificationRule(
                rule_id="test1",
                events=["tos:ObjectCreated:Post", "tos:ObjectCreated:Origin"],
                filter=NotificationFilter(
                    tos_key=NotificationFilterKey(
                        filter_rules=[
                            NotificationFilterRule(name="prefix", value="test-")
                        ]
                    )
                ),
                destination=NotificationDestination(
                    ve_faas=[DestinationVeFaaS(function_id=self.cloud_function)],
                    rocket_mq=[
                        DestinationRocketMQ(
                            role="trn:iam::{}:role/{}".format(self.account_id, self.mq_role_name),
                            instance_id=self.mq_instance_id,
                            topic="SDK",
                            access_key_id=self.mq_access_key_id
                        )
                    ]
                )
            )
        ]
        self.client.put_bucket_notification_type2(bucket_name, rules)
        out = self.client.get_bucket_notification_type2(bucket_name)

        self.assertTrue(out.version != '')
        self.assertEqual(len(out.rules), 1)
        self.assertEqual(out.rules[0].rule_id, rules[0].rule_id)
        self.assertEqual(out.rules[0].events, rules[0].events)
        self.assertEqual(out.rules[0].filter.tos_key.filter_rules[0].name, rules[0].filter.tos_key.filter_rules[0].name)
        self.assertEqual(out.rules[0].filter.tos_key.filter_rules[0].value,
                         rules[0].filter.tos_key.filter_rules[0].value)
        self.assertEqual(out.rules[0].destination.ve_faas[0].function_id, rules[0].destination.ve_faas[0].function_id)
        self.assertEqual(out.rules[0].destination.rocket_mq[0].role, rules[0].destination.rocket_mq[0].role)
        self.assertEqual(out.rules[0].destination.rocket_mq[0].topic, rules[0].destination.rocket_mq[0].topic)
        self.assertEqual(out.rules[0].destination.rocket_mq[0].access_key_id,
                         rules[0].destination.rocket_mq[0].access_key_id)
        self.assertEqual(out.rules[0].destination.rocket_mq[0].instance_id,
                         rules[0].destination.rocket_mq[0].instance_id)

    def retry_assert(self, func):
        for i in range(5):
            if func():
                return
            else:
                tim.sleep(i + 2)

        self.assertTrue(False)

    def test_delete_all(self):
        out = self.client2.list_buckets()
        task = TaskExecutor(15, clean_and_delete_bucket, None)
        for bucket in out.buckets:
            if bucket.name.startswith('sun'):
                if bucket.extranet_endpoint == self.endpoint2:
                    task.submit(self.client2, bucket.name)
                if bucket.extranet_endpoint == self.endpoint:
                    task.submit(self.client, bucket.name)
        task.run()


def _get_clean_endpoint(endpoint):
    if endpoint.startswith('http://'):
        return endpoint[7:]
    elif endpoint.startswith('https://'):
        return endpoint[8:]
    return endpoint


if __name__ == "__main__":
    unittest.main()



# # -*- coding: utf-8 -*-
# from tests.common import TosTestBase
#
# from tos.enum import (QueryOrderType, AggregationOperationType, SemanticQueryType,
#                       QueryOperationType)
#
# from tos.models2 import  AggregationRequest, QueryRequest
#
#
# class TestObject(TosTestBase):
#
#     def test_simple_query(self):
#         # base simple query
#         resp = self.client.simple_query(account_id=self.account_id,
#                                         dataset_name="windlike-2")
#         self.assertEqual(resp.status_code, 200)
#         self.assertEqual(100, len(resp.files))
#
#         # with sort,order,max_result
#         resp = self.client.simple_query(account_id=self.account_id,
#                                  dataset_name="windlike-2",
#                                  sort="FileModifiedTime",
#                                  order=QueryOrderType.ASC,
#                                         max_results=10)
#         self.assertEqual(resp.status_code, 200)
#         self.assertEqual(10, len(resp.files))
#         file_name = resp.files[0].file_name
#         # with next token
#         resp = self.client.simple_query(account_id=self.account_id,
#                                         dataset_name="windlike-2",
#                                         sort="FileModifiedTime",
#                                         order=QueryOrderType.ASC,
#                                         max_results=10,next_token=resp.next_token)
#         self.assertEqual(10, len(resp.files))
#         self.assertNotEquals(file_name , resp.files[0].file_name)
#         # with fields
#         resp = self.client.simple_query(account_id=self.account_id,
#                                         dataset_name="windlike-2",
#                                         sort="FileModifiedTime",
#                                         order=QueryOrderType.ASC,
#                                         max_results=10,with_fields=["TOSBucketName","FileName","ETag","TOSStorageClass",
#                                                                     "Size","ContentType","TOSCRC64","ServerSideEncryption",
#                                                                     "ServerSideEncryptionCustomerAlgorithm","TOSTaggingCount",
#                                                                     "TOSTagging","TOSUserMeta","TOSVersionId","TOSObjectType",
#                                                                     "TOSReplicationStatus","TOSIsDeleteMarker","AccountId"],
#                                         query=QueryRequest(
#                                             operation=QueryOperationType.AND,
#                                             sub_queries=[QueryRequest(operation=QueryOperationType.PREFIX,
#                                                                       field="FileName", value="test-py-sdk")])
#                                         )
#         self.assertEqual(1, len(resp.files))
#         self.assertEqual(True,resp.files[0].etag is not None)
#         # with query
#         resp = self.client.simple_query(account_id=self.account_id,
#                                         dataset_name="windlike-2",
#                                         sort="FileModifiedTime",
#                                         order=QueryOrderType.ASC,
#                                         max_results=10,  with_fields=["FileName"],
#                                         query=QueryRequest(
#                                             operation=QueryOperationType.AND,
#                                             sub_queries=[QueryRequest(operation=QueryOperationType.PREFIX,
#                                                                       field="FileName", value="1a5")])
#                                         )
#         self.assertEqual(0, len(resp.files))
#
#
#         # with aggregations
#
#         resp = self.client.simple_query(account_id=self.account_id,
#                                         dataset_name="windlike-2",
#                                         aggregations=[AggregationRequest("Size",AggregationOperationType.MAX),
#                                                       AggregationRequest("TOSStorageClass", AggregationOperationType.GROUP),
#                                                       AggregationRequest("ContentType", AggregationOperationType.GROUP)
#                                                       ])
#
#         self.assertEqual(resp.status_code, 200)
#
#     def test_semantic_query(self):
#         # 文字检索
#         resp = self.client.semantic_query(account_id=self.account_id,
#                                         dataset_name="img-test-only-incr",
#                                         semantic_query_type=SemanticQueryType.SemanticQueryTypeText,
#                                         max_results=10,
#                                         semantic_query_input="猫",
#                                         query=QueryRequest(
#                                             operation=QueryOperationType.AND,
#                                             sub_queries=[QueryRequest(operation=QueryOperationType.PREFIX,
#                                                                       field="FileName",value="1a5")]),
#                                         with_fields=["FileName"],
#                                           )
#
#         self.assertEqual(resp.status_code, 200)


import json
from typing import Dict

from .consts import LAST_MODIFY_TIME_DATE_FORMAT
from .models2 import Owner, RedirectAllRequestsTo, IndexDocument, ErrorDocument, RoutingRules, CustomDomainRule, \
    RealTimeLogConfiguration, RestoreJobParameters, BucketEncryptionRule, QueryOrderType, QueryRequest, \
    AggregationRequest, BucketInventoryConfiguration, SemanticQueryType, ReplicationRule, Tag,Rule
from .utils import check_enum_type
from typing import List



def to_complete_multipart_upload_request(parts: list):
    dic = {}
    if parts is None:
        return dic
    p = []
    for part in parts:
        p.append({'PartNumber': part.part_number, 'ETag': part.etag})
    dic['Parts'] = p
    return dic


def to_delete_multi_objects_request(objects: [], quiet: bool, recursive: bool):
    data = {}
    if objects:
        obs = []
        for o in objects:
            obs.append({'Key': o.key, 'VersionId': o.version_id})
        data['Objects'] = obs
    data['Quiet'] = quiet
    return data


def to_put_acl_request(owner: Owner, grants: []):
    data = {}
    if owner:
        data['Owner'] = {"ID": owner.id, "DisplayName": owner.display_name}
    if grants:
        convertor_grant = []
        for grant in grants:
            m = {}
            grantee = grant.grantee
            check_enum_type(grantee=grantee.type, canned=grantee.canned, permission=grant.permission)
            m['Grantee'] = {'ID': grantee.id, 'Type': grantee.type.value, 'DisplayName': grantee.display_name}
            if grantee.canned:
                m['Grantee']['Canned'] = grantee.canned.value

            m['Permission'] = grant.permission.value
            convertor_grant.append(m)
        data['Grants'] = convertor_grant

    return data


def to_put_bucket_cors_request(cors_rules: []):
    data = {}
    arr = []
    for cors_rule in cors_rules:
        info = {}
        if cors_rule.allowed_origins:
            info['AllowedOrigins'] = cors_rule.allowed_origins
        if cors_rule.allowed_methods:
            info['AllowedMethods'] = cors_rule.allowed_methods
        if cors_rule.allowed_headers:
            info['AllowedHeaders'] = cors_rule.allowed_headers
        if cors_rule.expose_headers:
            info['ExposeHeaders'] = cors_rule.expose_headers
        if cors_rule.max_age_seconds:
            info['MaxAgeSeconds'] = cors_rule.max_age_seconds
        if cors_rule.response_vary:
            info['ResponseVary'] = cors_rule.response_vary
        arr.append(info)

    data['CORSRules'] = arr

    return data


def to_put_bucket_mirror_back(rules: List[Rule]):
    rules = rules or []
    data = {}
    arr = []
    for rule in rules:
        info = {}
        if rule.id:
            info['ID'] = rule.id
        if rule.condition:
            info['Condition'] = {}
            if rule.condition.http_code:
                info['Condition']['HttpCode'] = rule.condition.http_code
            if rule.condition.key_prefix:
                info['Condition']['KeyPrefix'] = rule.condition.key_prefix
            if rule.condition.key_suffix:
                info['Condition']['KeySuffix'] = rule.condition.key_suffix
            if rule.condition.http_method:
                info['Condition']['HttpMethod'] = rule.condition.http_method
            if rule.condition.allow_host:
                info['Condition']['AllowHost'] = rule.condition.allow_host
        if rule.redirect:
            info['Redirect'] = {}
            if rule.redirect.redirect_type:
                info['Redirect']['RedirectType'] = rule.redirect.redirect_type.value
            if rule.redirect.fetch_source_on_redirect is not None:
                info['Redirect']['FetchSourceOnRedirect'] = rule.redirect.fetch_source_on_redirect
            if rule.redirect.fetch_source_on_redirect_with_query is not None:
                info['Redirect']['FetchSourceOnRedirectWithQuery'] = rule.redirect.fetch_source_on_redirect_with_query
            if rule.redirect.public_source:
                info['Redirect']['PublicSource'] = {}
                if rule.redirect.public_source.source_endpoint:
                    info['Redirect']['PublicSource']['SourceEndpoint'] = {}
                    if rule.redirect.public_source.source_endpoint.primary:
                        info['Redirect']['PublicSource']['SourceEndpoint'][
                            'Primary'] = rule.redirect.public_source.source_endpoint.primary
                    if rule.redirect.public_source.source_endpoint.follower:
                        info['Redirect']['PublicSource']['SourceEndpoint'][
                            'Follower'] = rule.redirect.public_source.source_endpoint.follower
                if rule.redirect.public_source.fixed_endpoint is not None:
                    info['Redirect']['PublicSource']['FixedEndpoint'] = rule.redirect.public_source.fixed_endpoint
            if rule.redirect.private_source:
                info['Redirect']['PrivateSource'] = rule.redirect.private_source.to_dict()
            if rule.redirect.pass_query is not None:
                info['Redirect']['PassQuery'] = rule.redirect.pass_query
            if rule.redirect.follow_redirect is not None:
                info['Redirect']['FollowRedirect'] = rule.redirect.follow_redirect
            if rule.redirect.mirror_header:
                info['Redirect']['MirrorHeader'] = {}
                if rule.redirect.mirror_header.pass_all:
                    info['Redirect']['MirrorHeader']['PassAll'] = rule.redirect.mirror_header.pass_all
                if rule.redirect.mirror_header.pass_headers:
                    info['Redirect']['MirrorHeader']['Pass'] = rule.redirect.mirror_header.pass_headers
                if rule.redirect.mirror_header.remove:
                    info['Redirect']['MirrorHeader']['Remove'] = rule.redirect.mirror_header.remove
                if rule.redirect.mirror_header.set_header:
                    info['Redirect']['MirrorHeader']['Set'] = []
                    for one_kv in rule.redirect.mirror_header.set_header:
                        kv = {
                            'Key': one_kv.key,
                            'Value': one_kv.value
                        }
                        info['Redirect']['MirrorHeader']['Set'].append(kv)
            if rule.redirect.transform:
                info['Redirect']['Transform'] = {}
                if rule.redirect.transform.with_key_prefix:
                    info['Redirect']['Transform']['WithKeyPrefix'] = rule.redirect.transform.with_key_prefix
                if rule.redirect.transform.with_key_suffix:
                    info['Redirect']['Transform']['WithKeySuffix'] = rule.redirect.transform.with_key_suffix
                if rule.redirect.transform.replace_key_prefix:
                    info['Redirect']['Transform']['ReplaceKeyPrefix'] = {}
                    if rule.redirect.transform.replace_key_prefix.key_prefix:
                        info['Redirect']['Transform']['ReplaceKeyPrefix'][
                            'KeyPrefix'] = rule.redirect.transform.replace_key_prefix.key_prefix
                    if rule.redirect.transform.replace_key_prefix.replace_with:
                        info['Redirect']['Transform']['ReplaceKeyPrefix'][
                            'ReplaceWith'] = rule.redirect.transform.replace_key_prefix.replace_with
            if rule.redirect.fetch_header_to_meta_data_rules:
                meta_data_rules = []
                for r in rule.redirect.fetch_header_to_meta_data_rules:
                    meta_data_rule = {
                        'SourceHeader': r.source_header,
                        'MetaDataSuffix': r.meta_data_suffix
                    }
                    meta_data_rules.append(meta_data_rule)
                info['Redirect']['FetchHeaderToMetaDataRules'] = meta_data_rules
        arr.append(info)

    data['Rules'] = arr

    return data


def to_put_bucket_lifecycle(rules: []):
    data = {}
    arr = []
    for rule in rules:
        info = {}

        if rule.id:
            info['ID'] = rule.id

        if rule.prefix:
            info['Prefix'] = rule.prefix

        if rule.status:
            info['Status'] = rule.status.value

        if rule.expiration:
            info['Expiration'] = {}
            if rule.expiration.days:
                info['Expiration']['Days'] = rule.expiration.days
            if rule.expiration.date:
                info['Expiration']['Date'] = rule.expiration.date.strftime(LAST_MODIFY_TIME_DATE_FORMAT)

        if rule.no_current_version_expiration:
            info['NoncurrentVersionExpiration'] = {}
            if rule.no_current_version_expiration.no_current_days:
                info['NoncurrentVersionExpiration'][
                    'NoncurrentDays'] = rule.no_current_version_expiration.no_current_days
            if rule.no_current_version_expiration.non_current_date:
                info['NoncurrentVersionExpiration'][
                    'NoncurrentDate'] = rule.no_current_version_expiration.non_current_date.strftime(
                    LAST_MODIFY_TIME_DATE_FORMAT)

        if rule.abort_in_complete_multipart_upload and rule.abort_in_complete_multipart_upload.days_after_init:
            info['AbortIncompleteMultipartUpload'] = {
                'DaysAfterInitiation': rule.abort_in_complete_multipart_upload.days_after_init
            }

        if rule.tags:
            tag_arr = []
            for t in rule.tags:
                tag_arr.append({
                    'Key': t.key,
                    'Value': t.value
                })
            info['Tags'] = tag_arr

        if rule.transitions:
            trans_arr = []
            for t in rule.transitions:
                transition_info = {}
                if t.days:
                    transition_info['Days'] = t.days
                if t.date:
                    transition_info['Date'] = t.date.strftime(LAST_MODIFY_TIME_DATE_FORMAT)
                if t.storage_class:
                    transition_info['StorageClass'] = t.storage_class.value
                trans_arr.append(transition_info)
            info['Transitions'] = trans_arr

        if rule.non_current_version_transitions:
            current_version_transition_arr = []
            for tr in rule.non_current_version_transitions:
                non_current_version_transition_info = {}
                if tr.non_current_days:
                    non_current_version_transition_info['NoncurrentDays'] = tr.non_current_days
                if tr.storage_class:
                    non_current_version_transition_info['StorageClass'] = tr.storage_class.value
                if tr.non_current_date:
                    non_current_version_transition_info['NoncurrentDate'] = tr.non_current_date.strftime(
                        LAST_MODIFY_TIME_DATE_FORMAT)

                current_version_transition_arr.append(non_current_version_transition_info)

            info['NoncurrentVersionTransitions'] = current_version_transition_arr

        if rule.filter:
            info['Filter'] = {}
            if rule.filter.object_size_greater_than:
                info['Filter']['ObjectSizeGreaterThan'] = rule.filter.object_size_greater_than
            if rule.filter.greater_than_include_equal.value:
                info['Filter']['GreaterThanIncludeEqual'] = rule.filter.greater_than_include_equal.value
            if rule.filter.object_size_less_than:
                info['Filter']['ObjectSizeLessThan'] = rule.filter.object_size_less_than
            if rule.filter.less_than_include_equal.value:
                info['Filter']['LessThanIncludeEqual'] = rule.filter.less_than_include_equal.value
        arr.append(info)

    data['Rules'] = arr
    return data


def to_put_tagging(tags: []):
    info = []
    for tag in tags:
        info.append({
            'Key': tag.key,
            'Value': tag.value
        })
    data = {'TagSet': {
        'Tags': info
    }}
    return data


def to_fetch_object(url: str, object: str = None, ignore_same_key=None, hex_md5=None, content_md5=None):
    info = {'URL': url}
    if object:
        info['Object'] = object
    if ignore_same_key:
        info['IgnoreSameKey'] = ignore_same_key
    if not content_md5 and hex_md5:
        content_md5 = hex_md5
    if content_md5:
        info['ContentMD5'] = content_md5

    return info


def to_put_fetch_object(url: str, object: str = None, ignore_same_key=None, hex_md5=None, content_md5=None,
                        callback_url=None, callback_host=None, callback_body=None, callback_body_type=None):
    info = to_fetch_object(url, object, ignore_same_key, hex_md5, content_md5, )
    if callback_url:
        info['CallbackUrl'] = callback_url
    if callback_host:
        info['CallbackHost'] = callback_host
    if callback_body:
        info['CallbackBody'] = callback_body
    if callback_body_type:
        info['CallbackBodyType'] = callback_body_type
    return info


def to_put_replication(role: str, rules:List[ReplicationRule]):
    rules = rules or []
    info = {}
    if role:
        info['Role'] = role
    r = []
    for rule in rules:
        data = {}
        if rule.id:
            data['ID'] = rule.id
        if rule.status:
            data['Status'] = rule.status.value
        if rule.prefix_set:
            data['PrefixSet'] = rule.prefix_set
        if rule.destination:
            data['Destination'] = {}
            if rule.destination.bucket:
                data['Destination']['Bucket'] = rule.destination.bucket
            if rule.destination.location:
                data['Destination']['Location'] = rule.destination.location
            if rule.destination.storage_class:
                data['Destination']['StorageClass'] = rule.destination.storage_class.value
            if rule.destination.storage_class_inherit_directive:
                data['Destination'][
                    'StorageClassInheritDirective'] = rule.destination.storage_class_inherit_directive.value
        if rule.historical_object_replication:
            data['HistoricalObjectReplication'] = rule.historical_object_replication.value
        if rule.progress:
            data['Progress'] = {}
            if rule.progress.historical_object:
                data['Progress']['HistoricalObject'] = rule.progress.historical_object
            if rule.progress.new_object:
                data['Progress']['NewObject'] = rule.progress.new_object
        if rule.tags:
            data['Tags'] = [Tag.to_dict(t) for t in rule.tags]
        if rule.access_control_translation and rule.access_control_translation.owner:
            data['AccessControlTranslation'] = {"Owner": rule.access_control_translation.owner}
        r.append(data)
    info['Rules'] = r
    return info


def to_put_bucket_website(redirect_all_requests_to: RedirectAllRequestsTo,
                          index_document: IndexDocument, error_document: ErrorDocument, routing_rules: RoutingRules):
    info = {}
    if redirect_all_requests_to:
        info['RedirectAllRequestsTo'] = {}
        if redirect_all_requests_to.host_name:
            info['RedirectAllRequestsTo']['HostName'] = redirect_all_requests_to.host_name
        if redirect_all_requests_to.protocol:
            info['RedirectAllRequestsTo']['Protocol'] = redirect_all_requests_to.protocol
    if index_document:
        info['IndexDocument'] = {}
        if index_document.suffix:
            info['IndexDocument']['Suffix'] = index_document.suffix
        if index_document.forbidden_sub_dir:
            info['IndexDocument']['ForbiddenSubDir'] = index_document.forbidden_sub_dir

    if error_document:
        info['ErrorDocument'] = {}
        if error_document.key:
            info['ErrorDocument']['Key'] = error_document.key

    if routing_rules:
        info['RoutingRules'] = []
        if routing_rules.rules and len(routing_rules.rules) > 0:
            for rule in routing_rules.rules:
                rule_mp = {}
                if rule.condition:
                    rule_mp['Condition'] = {}
                    if rule.condition.key_prefix_equals:
                        rule_mp['Condition']['KeyPrefixEquals'] = rule.condition.key_prefix_equals
                    if rule.condition.http_error_code_returned_equals:
                        rule_mp['Condition'][
                            'HttpErrorCodeReturnedEquals'] = rule.condition.http_error_code_returned_equals
                if rule.redirect:
                    rule_mp['Redirect'] = {}
                    if rule.redirect.host_name:
                        rule_mp['Redirect']['HostName'] = rule.redirect.host_name
                    if rule.redirect.http_redirect_code:
                        rule_mp['Redirect']['HttpRedirectCode'] = rule.redirect.http_redirect_code
                    if rule.redirect.protocol:
                        rule_mp['Redirect']['Protocol'] = rule.redirect.protocol.value
                    if rule.redirect.replace_key_prefix_with:
                        rule_mp['Redirect']['ReplaceKeyPrefixWith'] = rule.redirect.replace_key_prefix_with
                    if rule.redirect.replace_key_with:
                        rule_mp['Redirect']['ReplaceKeyWith'] = rule.redirect.replace_key_with

                info['RoutingRules'].append(rule_mp)
    return info


def to_put_bucket_notification(cloudFunctionConfigurations: [], rocketMQConfigurations: []):
    info = {}
    if cloudFunctionConfigurations:
        info['CloudFunctionConfigurations'] = []
        for cloudFunctionConfiguration in cloudFunctionConfigurations:
            config = {}
            if cloudFunctionConfiguration.events:
                config['Events'] = cloudFunctionConfiguration.events
            if cloudFunctionConfiguration.id:
                config['RuleId'] = cloudFunctionConfiguration.id
            if cloudFunctionConfiguration.cloud_function:
                config['CloudFunction'] = cloudFunctionConfiguration.cloud_function
            if cloudFunctionConfiguration.filter:
                config['Filter'] = _get_bucket_notification_filter_map(cloudFunctionConfiguration.filter)
            info['CloudFunctionConfigurations'].append(config)

    if rocketMQConfigurations:
        info['RocketMQConfigurations'] = []
        for rocketMQConfiguration in rocketMQConfigurations:
            config = {}
            if rocketMQConfiguration.events:
                config['Events'] = rocketMQConfiguration.events
            if rocketMQConfiguration.id:
                config['RuleId'] = rocketMQConfiguration.id
            if rocketMQConfiguration.role:
                config['Role'] = rocketMQConfiguration.role
            if rocketMQConfiguration.rocket_mq:
                config['RocketMQ'] = {}
                if rocketMQConfiguration.rocket_mq.topic:
                    config['RocketMQ']['Topic'] = rocketMQConfiguration.rocket_mq.topic
                if rocketMQConfiguration.rocket_mq.access_key_id:
                    config['RocketMQ']['AccessKeyId'] = rocketMQConfiguration.rocket_mq.access_key_id
                if rocketMQConfiguration.rocket_mq.instance_id:
                    config['RocketMQ']['InstanceID'] = rocketMQConfiguration.rocket_mq.instance_id
            if rocketMQConfiguration.filter:
                config['Filter'] = _get_bucket_notification_filter_map(rocketMQConfiguration.filter)
            info['RocketMQConfigurations'].append(config)
    return info


def _get_bucket_notification_filter_map(filter: {}):
    filter_mp = {}
    if filter.key:
        filter_mp['TOSKey'] = {}
        if filter.key.rules and len(filter.key.rules) >= 1:
            filter_mp['TOSKey']['FilterRules'] = []
            for rule in filter.key.rules:
                rule_mp = {}
                if rule.name:
                    rule_mp['Name'] = rule.name
                if rule.value:
                    rule_mp['Value'] = rule.value
                filter_mp['TOSKey']['FilterRules'].append(rule_mp)
    return filter_mp


def to_put_custom_domain(custom_domain_rule: CustomDomainRule):
    info = {}
    if custom_domain_rule:
        info['CustomDomainRule'] = {}
        if custom_domain_rule.domain:
            info['CustomDomainRule']['Domain'] = custom_domain_rule.domain
        if custom_domain_rule.cname:
            info['CustomDomainRule']['Cname'] = custom_domain_rule.cname
        if custom_domain_rule.cert_id:
            info['CustomDomainRule']['CertId'] = custom_domain_rule.cert_id
        if custom_domain_rule.cert_status:
            info['CustomDomainRule']['CertStatus'] = custom_domain_rule.cert_status.value
        if custom_domain_rule.forbidden:
            info['CustomDomainRule']['Forbidden'] = custom_domain_rule.forbidden
        if custom_domain_rule.forbidden_reason:
            info['CustomDomainRule']['ForbiddenReason'] = custom_domain_rule.forbidden_reason

    return info


def to_put_bucket_real_time_log(configuation: RealTimeLogConfiguration):
    info = {}
    if configuation:
        info['RealTimeLogConfiguration'] = {}
        if configuation.role:
            info['RealTimeLogConfiguration']['Role'] = configuation.role
        if configuation.configuration:
            info['RealTimeLogConfiguration']['AccessLogConfiguration'] = {}
            if configuation.configuration.use_service_topic:
                info['RealTimeLogConfiguration']['AccessLogConfiguration'][
                    'UseServiceTopic'] = configuation.configuration.use_service_topic
            if configuation.configuration.tls_topic_id:
                info['RealTimeLogConfiguration']['AccessLogConfiguration'][
                    'TLSTopicID'] = configuation.configuration.tls_topic_id
            if configuation.configuration.tls_project_id:
                info['RealTimeLogConfiguration']['AccessLogConfiguration'][
                    'TLSProjectID'] = configuation.configuration.tls_project_id
    return info


def to_restore_object(days: int, tier: RestoreJobParameters):
    info = {}
    if days:
        info['Days'] = days
    if tier and tier.tier:
        info['RestoreJobParameters'] = {"Tier": tier.tier.value}

    return info

def to_semantic_query(dataset_name:str = None,
                      semantic_query_input:str =None,
                      semantic_query_type:SemanticQueryType =None,
                       max_results:int=100,with_fields:List[str]=None,query:QueryRequest=None):
    info = {}
    if dataset_name:
        info['DatasetName'] = dataset_name
    if semantic_query_input:
        info['SemanticQueryInput'] = semantic_query_input
    if semantic_query_type:
        info['SemanticQueryType'] = semantic_query_type.value
    if max_results > 0:
        info['MaxResults'] = max_results
    if with_fields:
        info['WithFields'] = with_fields
    if query:
        info['Query'] = query.to_dict()

    return json.dumps(info)


def to_simple_query(dataset_name:str = None,
                     sort:str= None,
                     order:QueryOrderType= None,
                     max_results:int= 100,
                     next_token:str = None,
                     with_fields:List[str] = None,
                     query:QueryRequest = None,
                     aggregations: List[AggregationRequest] = None):
    info = {}
    if dataset_name:
        info['DatasetName'] = dataset_name
    if sort:
        info['Sort'] = sort
    if order:
        info['Order'] = order.value
    if max_results>0:
        info['MaxResults'] = max_results
    if next_token:
        info['NextToken'] = next_token
    if with_fields:
        info['WithFields'] = with_fields
    if query:
        info['Query'] = query.to_dict()
    if aggregations:
        info['Aggregations'] = [a.to_dict() for a in aggregations]

    return json.dumps(info)



def to_bucket_encrypt(rule: BucketEncryptionRule):
    info = {}
    if rule and rule.apply_server_side_encryption_by_default:
        info['Rule'] = {}
        info['Rule']['ApplyServerSideEncryptionByDefault'] = {
            'SSEAlgorithm': rule.apply_server_side_encryption_by_default.sse_algorithm,
            'KMSMasterKeyID': rule.apply_server_side_encryption_by_default.kms_master_key_id
        }
    return info


def to_put_bucket_notification_type2(rules: [], version: str):
    info = {}
    if version:
        info['Version'] = version
    if rules:
        info['Rules'] = []
        for rule in rules:
            config = {'RuleID': rule.rule_id}
            if rule.events:
                config['Events'] = rule.events
            if rule.filter and rule.filter.tos_key and rule.filter.tos_key.filter_rules:
                config['Filter'] = {'TOSKey': {}}
                config_filter_rules = []
                for filter_rule in rule.filter.tos_key.filter_rules:
                    config_filter_rules.append({
                        'Name': filter_rule.name,
                        'Value': filter_rule.value,
                    })
                config['Filter']['TOSKey'] = {'FilterRules': config_filter_rules}
            if rule.destination:
                config['Destination'] = {}
                if rule.destination.rocket_mq:
                    rocket_mqs = []
                    for r in rule.destination.rocket_mq:
                        rocket_mq = {}
                        if r.role:
                            rocket_mq['Role'] = r.role
                        if r.instance_id:
                            rocket_mq['InstanceID'] = r.instance_id
                        if r.topic:
                            rocket_mq['Topic'] = r.topic
                        if r.access_key_id:
                            rocket_mq['AccessKeyID'] = r.access_key_id
                        rocket_mqs.append(rocket_mq)
                    config['Destination']['RocketMQ'] = rocket_mqs
                if rule.destination.ve_faas:
                    ve_faas = []
                    for r in rule.destination.ve_faas:
                        ve_faas.append({
                            'FunctionID': r.function_id
                        })
                    config['Destination']['VeFaaS'] = ve_faas

            info['Rules'].append(config)
    return info
